@@ -1,8 +1,9 @@
-from typing import Tuple
+from typing import Tuple, Optional, Dict
 import logging
 from collections import defaultdict, OrderedDict
 
 import pyvex
+import claripy
 from archinfo.arch_arm import is_arm_arch
 
 from ....concretization_strategies import SimConcretizationStrategyAny
@@ -130,6 +131,22 @@ class JumpTableProcessorState:
         self.regs_to_initialize = [ ]  # registers that we should initialize
 
 
+class RegOffsetAnnotation(claripy.Annotation):
+
+    __slots__ = ('reg_offset', )
+
+    def __init__(self, reg_offset: RegisterOffset):
+        self.reg_offset = reg_offset
+
+    @property
+    def relocatable(self):
+        return False
+
+    @property
+    def eliminatable(self):
+        return False
+
+
 class JumpTableProcessor(
     SimEngineLightVEXMixin,
     SimEngineLight,
@@ -159,6 +176,77 @@ class JumpTableProcessor(
         self._bp_sp_diff = bp_sp_diff  # bp - sp
         self._tsrc = set()  # a scratch variable to store source information for values
 
+        self._SPOFFSET_BASE = claripy.BVS('SpOffset', self.project.arch.bits, explicit_name=True)
+        self._REGOFFSET_BASE: Dict[int,claripy.ast.BV] = {}
+
+    @staticmethod
+    def _top(size: int):
+        return None
+
+    @staticmethod
+    def _is_top(expr) -> bool:
+        return expr is None
+
+    @staticmethod
+    def _is_spoffset(expr) -> bool:
+        return 'SpOffset' in expr.variables
+
+    def _get_spoffset_expr(self, sp_offset: SpOffset) -> claripy.ast.BV:
+        v = self._SPOFFSET_BASE.annotate(RegOffsetAnnotation(sp_offset))
+        return v
+
+    @staticmethod
+    def _extract_spoffset_from_expr(expr: claripy.ast.Base) -> Optional[SpOffset]:
+        if expr.op == "BVS":
+            for anno in expr.annotations:
+                if isinstance(anno, RegOffsetAnnotation):
+                    return anno.reg_offset
+        elif expr.op == "__add__":
+            if len(expr.args) == 1:
+                return JumpTableProcessor._extract_spoffset_from_expr(expr.args[0])
+            elif len(expr.args) == 2 and expr.args[1].op == "BVV":
+                sp_offset = JumpTableProcessor._extract_spoffset_from_expr(expr.args[0])
+                if sp_offset is not None:
+                    delta = expr.args[1]._model_concrete.value
+                    sp_offset += delta
+                    return sp_offset
+        elif expr.op == "__and__":
+            if len(expr.args) == 2 and expr.args[1].op == "BVV":
+                # ignore all masking on SpOffsets
+                return JumpTableProcessor._extract_spoffset_from_expr(expr.args[0])
+        return None
+
+    @staticmethod
+    def _is_registeroffset(expr) -> bool:
+        return 'RegisterOffset' in expr.variables
+
+    def _get_regoffset_expr(self, reg_offset: RegisterOffset, bits: int) -> claripy.ast.BV:
+        if bits not in self._REGOFFSET_BASE:
+            self._REGOFFSET_BASE[bits] = claripy.BVS('RegisterOffset', bits, explicit_name=True)
+        v = self._REGOFFSET_BASE[bits].annotate(RegOffsetAnnotation(reg_offset))
+        return v
+
+    @staticmethod
+    def _extract_regoffset_from_expr(expr: claripy.ast.Base) -> Optional[RegisterOffset]:
+        if expr.op == "BVS":
+            for anno in expr.annotations:
+                if isinstance(anno, RegOffsetAnnotation):
+                    return anno.reg_offset
+        elif expr.op == "__add__":
+            if len(expr.args) == 1:
+                return JumpTableProcessor._extract_regoffset_from_expr(expr.args[0])
+            elif len(expr.args) == 2 and expr.args[1].op == "BVV":
+                reg_offset = JumpTableProcessor._extract_regoffset_from_expr(expr.args[0])
+                if reg_offset is not None:
+                    delta = expr.args[1]._model_concrete.value
+                    reg_offset += delta
+                    return reg_offset
+        elif expr.op == "__and__":
+            if len(expr.args) == 2 and expr.args[1].op == "BVV":
+                # ignore all masking on SpOffsets
+                return JumpTableProcessor._extract_spoffset_from_expr(expr.args[0])
+        return None
+
     def _handle_WrTmp(self, stmt):
         self._tsrc = set()
         super()._handle_WrTmp(stmt)
@@ -195,19 +283,32 @@ class JumpTableProcessor(
 
     def _handle_Get(self, expr):
         if expr.offset == self.arch.bp_offset:
-            return SpOffset(self.arch.bits, self._bp_sp_diff)
+            return self._get_spoffset_expr(SpOffset(self.arch.bits, self._bp_sp_diff))
         elif expr.offset == self.arch.sp_offset:
-            return SpOffset(self.arch.bits, 0)
+            return self._get_spoffset_expr(SpOffset(self.arch.bits, 0))
         else:
             if expr.offset in self.state._registers:
                 self._tsrc |= set(self.state._registers[expr.offset][0])
-                return self.state._registers[expr.offset][1]
-            # the register does not exist
-            # we initialize it here
-            v = RegisterOffset(expr.result_size(self.tyenv), expr.offset, 0)
-            src = (self.block.addr, self.stmt_idx)
-            self._tsrc.add(src)
-            self.state._registers[expr.offset] = ([src], v)
+                v = self.state._registers[expr.offset][1]
+            else:
+                # the register does not exist
+                # we initialize it here
+                v = RegisterOffset(expr.result_size(self.tyenv), expr.offset, 0)
+                v = self._get_regoffset_expr(v, expr.result_size(self.tyenv))
+                src = (self.block.addr, self.stmt_idx)
+                self._tsrc.add(src)
+                self.state._registers[expr.offset] = ([src], v)
+
+            # make sure the size matches
+            # note that this is sometimes incorrect. for example, we do not differentiate between reads at ah and al...
+            # but it should be good enough for now (without switching state._registers to a real SimMemory, which will
+            # surely slow down stuff quite a bit)
+            if v is not None:
+                bits = expr.result_size(self.tyenv)
+                if v.size() > bits:
+                    v = v[bits - 1:0]
+                elif v.size() < bits:
+                    v = claripy.ZeroExt(bits - v.size(), v)
             return v
 
     def _handle_function(self, expr):  # pylint:disable=unused-argument,no-self-use
@@ -352,10 +453,11 @@ class JumpTableProcessor(
         if addr is None:
             return None
 
-        if isinstance(addr, SpOffset):
-            if addr.offset in self.state._stack:
-                self._tsrc = { self.state._stack[addr.offset][0] }
-                return self.state._stack[addr.offset][1]
+        if self._is_spoffset(addr):
+            spoffset = self._extract_spoffset_from_expr(addr)
+            if spoffset is not None and spoffset.offset in self.state._stack:
+                self._tsrc = { self.state._stack[spoffset.offset][0] }
+                return self.state._stack[spoffset.offset][1]
         elif isinstance(addr, int):
             # Load data from memory if it is mapped
             try:
@@ -363,16 +465,17 @@ class JumpTableProcessor(
                 return v
             except KeyError:
                 return None
-        elif isinstance(addr, RegisterOffset):
+        elif self._is_registeroffset(addr):
             # Load data from a register, but this register hasn't been initialized at this point
             # We will need to initialize this register during slice execution later
 
             # Try to get where this register is first accessed
-            if addr.reg in self.state._registers:
+            reg_offset = self._extract_regoffset_from_expr(addr)
+            if reg_offset is not None and reg_offset.reg in self.state._registers:
                 try:
-                    source = next(iter(src for src in self.state._registers[addr.reg][0] if src != 'const'))
+                    source = next(iter(src for src in self.state._registers[reg_offset.reg][0] if src != 'const'))
                     assert isinstance(source, tuple)
-                    self.state.regs_to_initialize.append(source + (addr.reg, addr.bits))
+                    self.state.regs_to_initialize.append(source + (reg_offset.reg, reg_offset.bits))
                 except StopIteration:
                     # we don't need to initialize this register
                     # it might be caused by an incorrect analysis result
@@ -465,7 +568,8 @@ class BSSHook:
             # it was never written to before. we overwrite it with unconstrained bytes
             for i in range(0, concrete_read_length, self.project.arch.bytes):
                 state.memory.store(concrete_read_addr + i, state.solver.Unconstrained('unconstrained',
-                                                                                      self.project.arch.bits))
+                                                                                      self.project.arch.bits),
+                                   endness=self.project.arch.memory_endness)
 
                 # job done :-)
 
@@ -712,8 +816,13 @@ class JumpTableResolver(IndirectJumpResolver):
                     if ret is None:
                         # Try the next state
                         continue
-                    jump_table, jumptable_addr, entry_size, jumptable_size, all_targets = ret
-                    ij_type = IndirectJumpType.Jumptable_AddressLoadedFromMemory
+                    jump_table, jumptable_addr, entry_size, jumptable_size, all_targets, sort = ret
+                    if sort == "jumptable":
+                        ij_type = IndirectJumpType.Jumptable_AddressLoadedFromMemory
+                    elif sort == "vtable":
+                        ij_type = IndirectJumpType.Vtable
+                    else:
+                        ij_type = IndirectJumpType.Unknown
                 elif ite_stmt is not None:
                     ret = self._try_resolve_targets_ite(r, addr, cfg, annotatedcfg, ite_stmt)
                     if ret is None:
@@ -734,7 +843,10 @@ class JumpTableResolver(IndirectJumpResolver):
                 ij: IndirectJump = cfg.indirect_jumps[addr]
                 if len(all_targets) > 1:
                     # It can be considered a jump table only if there are more than one jump target
-                    ij.jumptable = True
+                    if ij_type in (IndirectJumpType.Jumptable_AddressComputed, IndirectJumpType.Jumptable_AddressLoadedFromMemory):
+                        ij.jumptable = True
+                    else:
+                        ij.jumptable = False
                     ij.jumptable_addr = jumptable_addr
                     ij.jumptable_size = jumptable_size
                     ij.jumptable_entry_size = entry_size
@@ -1143,7 +1255,7 @@ class JumpTableResolver(IndirectJumpResolver):
     def _try_resolve_targets_load(self, r, addr, cfg, annotatedcfg, load_stmt, load_size, stmts_adding_base_addr,
                                   all_addr_holders):
         """
-        Try loading all jump targets from a jump table.
+        Try loading all jump targets from a jump table or a vtable.
         """
 
         # shorthand
@@ -1212,7 +1324,23 @@ class JumpTableResolver(IndirectJumpResolver):
                 jump_base_addr.base_addr = state.solver.eval(state.scratch.temps[jump_base_addr.tmp_1])
 
         all_targets = [ ]
-        total_cases = jumptable_addr._model_vsa.cardinality
+        jumptable_addr_vsa = jumptable_addr._model_vsa
+
+        if not isinstance(jumptable_addr_vsa, claripy.vsa.StridedInterval):
+            return None
+
+        # we may resolve a vtable (in C, e.g., the IO_JUMPS_FUNC in libc), but the stride of this load is usually 1
+        # while the read statement reads a word size at a time.
+        # we use this to differentiate between traditional jump tables (where each entry is some blocks that belong to
+        # the current function) and vtables (where each entry is a function).
+        if jumptable_addr_vsa.stride < load_size:
+            stride = load_size
+            total_cases = jumptable_addr_vsa.cardinality // load_size
+            sort = 'vtable'  # it's probably a vtable!
+        else:
+            stride = jumptable_addr_vsa.stride
+            total_cases = jumptable_addr_vsa.cardinality
+            sort = 'jumptable'
 
         if total_cases > self._max_targets:
             # We resolved too many targets for this indirect jump. Something might have gone wrong.
@@ -1245,9 +1373,11 @@ class JumpTableResolver(IndirectJumpResolver):
 
         # Load the jump table from memory
         should_skip = False
-        for idx, a in enumerate(state.solver.eval_upto(jumptable_addr, total_cases)):
+        for idx, a in enumerate(range(min_jumptable_addr, max_jumptable_addr + 1, stride)):
             if idx % 100 == 0 and idx != 0:
                 l.debug("%d targets have been resolved for the indirect jump at %#x...", idx, addr)
+            if idx >= total_cases:
+                break
             target = cfg._fast_memory_load_pointer(a, size=load_size)
             if target is None:
                 l.debug("Cannot load pointer from address %#x. Skip.", a)
@@ -1317,7 +1447,7 @@ class JumpTableResolver(IndirectJumpResolver):
         if illegal_target_found:
             return None
 
-        return jump_table, min_jumptable_addr, load_size, total_cases * load_size, all_targets
+        return jump_table, min_jumptable_addr, load_size, total_cases * load_size, all_targets, sort
 
     def _try_resolve_targets_ite(self, r, addr, cfg, annotatedcfg, ite_stmt: pyvex.IRStmt.WrTmp):  # pylint:disable=unused-argument
         """
@@ -1381,25 +1511,25 @@ class JumpTableResolver(IndirectJumpResolver):
             if sort == 'mem_write':
                 bp = BP(when=BP_BEFORE, enabled=True, action=StoreHook.hook,
                         condition=lambda _s, a=block_addr, idx=stmt_idx:
-                            _s.scratch.bbl_addr == a and _s.inspect.statement == idx
+                            _s.scratch.bbl_addr == a and _s.scratch.stmt_idx == idx
                         )
                 state.inspect.add_breakpoint('mem_write', bp)
             elif sort == 'mem_read':
                 hook = LoadHook()
                 bp0 = BP(when=BP_BEFORE, enabled=True, action=hook.hook_before,
                          condition=lambda _s, a=block_addr, idx=stmt_idx:
-                            _s.scratch.bbl_addr == a and _s.inspect.statement == idx
+                            _s.scratch.bbl_addr == a and _s.scratch.stmt_idx == idx
                          )
                 state.inspect.add_breakpoint('mem_read', bp0)
                 bp1 = BP(when=BP_AFTER, enabled=True, action=hook.hook_after,
                          condition=lambda _s, a=block_addr, idx=stmt_idx:
-                            _s.scratch.bbl_addr == a and _s.inspect.statement == idx
+                            _s.scratch.bbl_addr == a and _s.scratch.stmt_idx == idx
                          )
                 state.inspect.add_breakpoint('mem_read', bp1)
             elif sort == 'reg_write':
                 bp = BP(when=BP_BEFORE, enabled=True, action=PutHook.hook,
                         condition=lambda _s, a=block_addr, idx=stmt_idx:
-                            _s.scratch.bbl_addr == a and _s.inspect.statement == idx
+                            _s.scratch.bbl_addr == a and _s.scratch.stmt_idx == idx
                         )
                 state.inspect.add_breakpoint('reg_write', bp)
             else:
