@@ -1,10 +1,11 @@
-from typing import Set, Dict, List
+from typing import Set, Dict, List, Tuple, Any, Optional
 from collections import defaultdict
 
 from ailment.block import Block
 from ailment.statement import Statement, Assignment, Store, Call
 from ailment.expression import Register, Convert, Load, StackBaseOffset
 
+from ...knowledge_plugins.key_definitions.constants import OP_AFTER
 from ...code_location import CodeLocation
 from ...analyses.reaching_definitions.external_codeloc import ExternalCodeLocation
 from ...sim_variable import SimStackVariable
@@ -13,6 +14,7 @@ from ...knowledge_plugins.key_definitions import atoms
 from ...knowledge_plugins.key_definitions.definition import Definition
 from .. import Analysis, AnalysesHub
 from .ailblock_walker import AILBlockWalker
+from .ailgraph_walker import AILGraphWalker
 from .block_simplifier import BlockSimplifier
 
 
@@ -24,11 +26,11 @@ class AILSimplifier(Analysis):
     """
     Perform function-level simplifications.
     """
-    def __init__(self, func, func_graph=None, remove_dead_memdefs=False, unify_variables=False,
-                 reaching_definitions=None):
+    def __init__(self, func, func_graph=None, remove_dead_memdefs=False, unify_variables=False):
         self.func = func
         self.func_graph = func_graph if func_graph is not None else func.graph
-        self._reaching_definitions = reaching_definitions
+        self._reaching_definitions = None
+        self._propagator = None
 
         self._remove_dead_memdefs = remove_dead_memdefs
         self._unify_vars = unify_variables
@@ -36,20 +38,73 @@ class AILSimplifier(Analysis):
         self._calls_to_remove: Set[CodeLocation] = set()
         self.blocks = {}  # Mapping nodes to simplified blocks
 
+        self.simplified: bool = False
         self._simplify()
 
     def _simplify(self):
 
-        if self._unify_vars:
-            self._unify_local_variables()
-            self._fold_call_exprs()
-
         folded_exprs = self._fold_exprs()
+        self.simplified |= folded_exprs
         if folded_exprs:
+            self._rebuild_func_graph()
             # reading definition analysis results are no longer reliable
             return
 
-        self._remove_dead_assignments()
+        if self._unify_vars:
+            r = self._unify_local_variables()
+            if r:
+                self.simplified = True
+                self._rebuild_func_graph()
+            # _fold_call_exprs() may set self._calls_to_remove, which will be honored in _remove_dead_assignments()
+            r = self._fold_call_exprs()
+            if r:
+                self.simplified = True
+                self._rebuild_func_graph()
+
+        r = self._remove_dead_assignments()
+        if r:
+            self.simplified = True
+            self._rebuild_func_graph()
+
+    def _rebuild_func_graph(self):
+        def _handler(node):
+            return self.blocks.get(node, None)
+        AILGraphWalker(self.func_graph, _handler, replace_nodes=True).walk()
+        self.blocks = {}
+
+    def _compute_reaching_definitions(self):
+        # Computing reaching definitions or return the cached one
+        if self._reaching_definitions is not None:
+            return self._reaching_definitions
+        rd = self.project.analyses.ReachingDefinitions(subject=self.func, func_graph=self.func_graph,
+                                                       observe_callback=self._simplify_function_rd_observe_callback)
+        self._reaching_definitions = rd
+        return rd
+
+    @staticmethod
+    def _simplify_function_rd_observe_callback(ob_type, **kwargs):
+        if ob_type != 'node':
+            return False
+        op_type = kwargs.pop('op_type')
+        return op_type == OP_AFTER
+
+    def _compute_propagation(self):
+        # Propagate expressions or return the existing result
+        if self._propagator is not None:
+            return self._propagator
+        prop = self.project.analyses.Propagator(func=self.func, func_graph=self.func_graph)
+        self._propagator = prop
+        return prop
+
+    def _clear_cache(self) -> None:
+        self._propagator = None
+        self._reaching_definitions = None
+
+    def _clear_propagator_cache(self) -> None:
+        self._propagator = None
+
+    def _clear_reaching_definitions_cache(self) -> None:
+        self._reaching_definitions = None
 
     def _fold_exprs(self):
         """
@@ -57,41 +112,56 @@ class AILSimplifier(Analysis):
         """
 
         # propagator
-        propagator = self.project.analyses.Propagator(func=self.func, func_graph=self.func_graph)
-        replacements = list(propagator._states.values())[0]._replacements
+        propagator = self._compute_propagation()
+        replacements = propagator.replacements
 
         # take replacements and rebuild the corresponding blocks
-        replacements_by_block_addrs = defaultdict(dict)
+        replacements_by_block_addrs_and_idx = defaultdict(dict)
         for codeloc, reps in replacements.items():
             if reps:
-                replacements_by_block_addrs[codeloc.block_addr][codeloc] = reps
+                replacements_by_block_addrs_and_idx[(codeloc.block_addr, codeloc.block_idx)][codeloc] = reps
 
-        if not replacements_by_block_addrs:
+        if not replacements_by_block_addrs_and_idx:
             return False
 
-        blocks_by_addr = dict((node.addr, node) for node in self.func_graph.nodes())
+        blocks_by_addr_and_idx = dict(((node.addr, node.idx), node) for node in self.func_graph.nodes())
 
-        for block_addr, reps in replacements_by_block_addrs.items():
-            block = blocks_by_addr[block_addr]
-            new_block = BlockSimplifier._replace_and_build(block, reps)
+        replaced = False
+        for (block_addr, block_idx), reps in replacements_by_block_addrs_and_idx.items():
+            block = blocks_by_addr_and_idx[(block_addr, block_idx)]
+            r, new_block = BlockSimplifier._replace_and_build(block, reps)
+            replaced |= r
             self.blocks[block] = new_block
 
-        return True
+        if replaced:
+            # blocks have been rebuilt - expression propagation results are no longer reliable
+            self._clear_cache()
+        return replaced
 
-    def _unify_local_variables(self):
+    def _unify_local_variables(self) -> bool:
         """
         Find variables that are definitely equivalent and then eliminate the unnecessary copies.
         """
-        prop = self.project.analyses.Propagator(func=self.func, func_graph=self.func_graph)
+
+        simplified = False
+
+        prop = self._compute_propagation()
         if not prop.equivalence:
-            return
+            return simplified
 
-        addr2block: Dict[int, Block] = { }
+        addr_and_idx_to_block: Dict[Tuple[int,int], Block] = { }
         for block in self.func_graph.nodes():
-            addr2block[block.addr] = block
+            addr_and_idx_to_block[(block.addr, block.idx)] = block
 
+        equivalences: Dict[Any,Set[Equivalence]] = defaultdict(set)
         for eq in prop.equivalence:
-            eq: Equivalence
+            equivalences[eq.atom1].add(eq)
+
+        for _, eqs in equivalences.items():
+            if len(eqs) > 1:
+                continue
+
+            eq = next(iter(eqs))
 
             # Acceptable equivalence classes:
             #
@@ -121,7 +191,8 @@ class AILSimplifier(Analysis):
                 continue
 
             # find the definition of this register
-            defs = self._reaching_definitions.all_uses.get_uses_by_location(eq.codeloc)
+            rd = self._compute_reaching_definitions()
+            defs = rd.all_uses.get_uses_by_location(eq.codeloc)
             for def_ in defs:
                 def_: Definition
                 if isinstance(def_.atom, atoms.Register) and def_.atom.reg_offset == reg.reg_offset:
@@ -135,7 +206,8 @@ class AILSimplifier(Analysis):
                 continue
 
             # find all uses of this definition
-            all_uses: Set[CodeLocation] = self._reaching_definitions.all_uses.get_uses(the_def)
+            # we make a copy of the set since we may touch the set (uses) when replacing expressions
+            all_uses: Set[CodeLocation] = set(rd.all_uses.get_uses(the_def))
 
             # TODO: We can only replace all these uses with the stack variable if the stack variable isn't
             # TODO: re-assigned of a new value. Perform this check.
@@ -145,7 +217,7 @@ class AILSimplifier(Analysis):
                 if u == eq.codeloc:
                     # skip the very initial assignment location
                     continue
-                old_block = addr2block.get(u.block_addr, None)
+                old_block = addr_and_idx_to_block.get((u.block_addr, u.block_idx), None)
                 if old_block is None:
                     continue
 
@@ -162,9 +234,16 @@ class AILSimplifier(Analysis):
                 else:
                     raise RuntimeError("Unsupported atom0 type %s." % type(eq.atom0))
 
-                self._replace_expr_and_update_block(the_block, u.stmt_idx, stmt, the_def, eq.atom1, dst)
+                r, new_block = self._replace_expr_and_update_block(the_block, u.stmt_idx, stmt, the_def, u, eq.atom1,
+                                                                   dst)
+                if r:
+                    self.blocks[old_block] = new_block
+                simplified |= r
 
-    def _fold_call_exprs(self):
+        # no need to clear cache at the end of this function
+        return simplified
+
+    def _fold_call_exprs(self) -> bool:
         """
         Fold a call expression (statement) into other statements if the return value of the call expression (statement)
         is only used once, and the use site and the call site belongs to the same supernode.
@@ -179,13 +258,15 @@ class AILSimplifier(Analysis):
             if (func()) ...
         """
 
-        prop = self.project.analyses.Propagator(func=self.func, func_graph=self.func_graph)
-        if not prop.equivalence:
-            return
+        simplified = False
 
-        addr2block: Dict[int, Block] = { }
+        prop = self._compute_propagation()
+        if not prop.equivalence:
+            return simplified
+
+        addr_and_idx_to_block: Dict[Tuple[int,int], Block] = { }
         for block in self.func_graph.nodes():
-            addr2block[block.addr] = block
+            addr_and_idx_to_block[(block.addr, block.idx)] = block
 
         for eq in prop.equivalence:
             eq: Equivalence
@@ -199,7 +280,8 @@ class AILSimplifier(Analysis):
                     continue
 
                 # find the definition of this register
-                defs = [ d for d in self._reaching_definitions.all_definitions
+                rd = self._compute_reaching_definitions()
+                defs = [ d for d in rd.all_definitions
                          if d.codeloc == eq.codeloc
                          and isinstance(d.atom, atoms.Register)
                          and d.atom.reg_offset == eq.atom0.reg_offset
@@ -209,19 +291,21 @@ class AILSimplifier(Analysis):
                 the_def: Definition = defs[0]
 
                 # find all uses of this definition
-                all_uses: Set[CodeLocation] = self._reaching_definitions.all_uses.get_uses(the_def)
+                all_uses: Set[CodeLocation] = set(rd.all_uses.get_uses(the_def))
 
                 if len(all_uses) != 1:
                     continue
                 u = next(iter(all_uses))
 
                 # check if the use and the definition is within the same supernode
-                super_node_blocks = self._get_super_node_blocks(addr2block[the_def.codeloc.block_addr])
+                super_node_blocks = self._get_super_node_blocks(
+                    addr_and_idx_to_block[(the_def.codeloc.block_addr, the_def.codeloc.block_idx)]
+                )
                 if u.block_addr not in set(b.addr for b in super_node_blocks):
                     continue
 
                 # replace all uses
-                old_block = addr2block.get(u.block_addr, None)
+                old_block = addr_and_idx_to_block.get((u.block_addr, u.block_idx), None)
                 if old_block is None:
                     continue
 
@@ -235,11 +319,17 @@ class AILSimplifier(Analysis):
                 else:
                     continue
 
-                replaced = self._replace_expr_and_update_block(the_block, u.stmt_idx, stmt, the_def, src, dst)
+                replaced, new_block = self._replace_expr_and_update_block(the_block, u.stmt_idx, stmt, the_def, u, src,
+                                                                          dst)
 
                 if replaced:
+                    self.blocks[old_block] = new_block
                     # this call has been folded to the use site. we can remove this call.
                     self._calls_to_remove.add(eq.codeloc)
+                    simplified = True
+
+        # no need to clear the cache at the end of this method
+        return simplified
 
     def _get_super_node_blocks(self, start_node: Block) -> List[Block]:
 
@@ -262,27 +352,29 @@ class AILSimplifier(Analysis):
                 break
         return lst
 
-    def _replace_expr_and_update_block(self, block, stmt_idx, stmt, the_def, src_expr, dst_expr) -> bool:
+    def _replace_expr_and_update_block(self, block, stmt_idx, stmt, the_def, codeloc, src_expr,
+                                       dst_expr) -> Tuple[bool,Optional[Block]]:
         replaced, new_stmt = stmt.replace(src_expr, dst_expr)
         if replaced:
             new_block = block.copy()
             new_block.statements = block.statements[::]
             new_block.statements[stmt_idx] = new_stmt
-            self.blocks[block] = new_block
 
             # update the uses
-            self._reaching_definitions.all_uses.remove_uses(the_def)
-            return True
+            rd = self._compute_reaching_definitions()
+            rd.all_uses.remove_use(the_def, codeloc)
+            return True, new_block
 
-        return False
+        return False, None
 
-    def _remove_dead_assignments(self):
+    def _remove_dead_assignments(self) -> bool:
 
-        stmts_to_remove_per_block = defaultdict(set)
+        stmts_to_remove_per_block: Dict[Tuple[int,int],Set[int]] = defaultdict(set)
 
         # Find all statements that should be removed
 
-        for def_ in self._reaching_definitions.all_definitions:  # type: Definition
+        rd = self._compute_reaching_definitions()
+        for def_ in rd.all_definitions:  # type: Definition
             if def_.dummy:
                 continue
             # we do not remove references to global memory regions no matter what
@@ -290,10 +382,16 @@ class AILSimplifier(Analysis):
                 continue
             if not self._remove_dead_memdefs and isinstance(def_.atom, atoms.MemoryLocation):
                 continue
-            uses = self._reaching_definitions.all_uses.get_uses(def_)
+            uses = rd.all_uses.get_uses(def_)
 
             if not uses:
-                stmts_to_remove_per_block[def_.codeloc.block_addr].add(def_.codeloc.stmt_idx)
+                stmts_to_remove_per_block[(def_.codeloc.block_addr, def_.codeloc.block_idx)].add(def_.codeloc.stmt_idx)
+
+        for codeloc in self._calls_to_remove:
+            # this call can be removed. make sure it exists in stmts_to_remove_per_block
+            stmts_to_remove_per_block[codeloc.block_addr, codeloc.block_idx].add(codeloc.stmt_idx)
+
+        simplified = False
 
         # Remove the statements
         for old_block in self.func_graph.nodes():
@@ -304,11 +402,14 @@ class AILSimplifier(Analysis):
             if not isinstance(block, Block):
                 continue
 
-            if block.addr not in stmts_to_remove_per_block:
+            if (block.addr, block.idx) not in stmts_to_remove_per_block:
                 continue
 
             new_statements = [ ]
-            stmts_to_remove = stmts_to_remove_per_block[block.addr]
+            stmts_to_remove = stmts_to_remove_per_block[(block.addr, block.idx)]
+
+            if not stmts_to_remove:
+                continue
 
             for idx, stmt in enumerate(block.statements):
                 if idx in stmts_to_remove:
@@ -316,15 +417,20 @@ class AILSimplifier(Analysis):
                         # Skip Assignment and Store statements
                         # if this statement triggers a call, it should not be removed
                         if not self._statement_has_call_exprs(stmt):
+                            simplified = True
                             continue
                     elif isinstance(stmt, Call):
                         codeloc = CodeLocation(block.addr, idx, ins_addr=stmt.ins_addr)
                         if codeloc in self._calls_to_remove:
                             # this call can be removed
+                            simplified = True
                             continue
-                        # the return expr is not used. it should not have return expr
-                        stmt = stmt.copy()
-                        stmt.ret_expr = None
+
+                        if stmt.ret_expr is not None:
+                            # the return expr is not used. it should not have return expr
+                            stmt = stmt.copy()
+                            stmt.ret_expr = None
+                            simplified = True
                     else:
                         # Should not happen!
                         raise NotImplementedError()
@@ -335,6 +441,8 @@ class AILSimplifier(Analysis):
             new_block.statements = new_statements
             self.blocks[old_block] = new_block
 
+        return simplified
+
     @staticmethod
     def _statement_has_call_exprs(stmt: Statement) -> bool:
 
