@@ -1,6 +1,6 @@
 from collections import defaultdict
 import logging
-from typing import List, Optional, Iterable, Union, Type, TYPE_CHECKING
+from typing import Dict, List, Tuple, Set, Optional, Iterable, Union, Type, TYPE_CHECKING
 
 import networkx
 
@@ -12,14 +12,14 @@ from ...utils import timethis
 from ...calling_conventions import SimRegArg, SimStackArg, SimFunctionArgument
 from ...sim_type import SimTypeChar, SimTypeInt, SimTypeLongLong, SimTypeShort, SimTypeFunction, SimTypeBottom
 from ...sim_variable import SimVariable, SimStackVariable, SimRegisterVariable
-from ...knowledge_plugins.key_definitions.constants import OP_BEFORE, OP_AFTER
+from ...knowledge_plugins.key_definitions.constants import OP_BEFORE
 from ...procedures.stubs.UnresolvableCallTarget import UnresolvableCallTarget
 from ...procedures.stubs.UnresolvableJumpTarget import UnresolvableJumpTarget
 from .. import Analysis, register_analysis
 from ..cfg.cfg_base import CFGBase
 from .ailgraph_walker import AILGraphWalker
 from .ailblock_walker import AILBlockWalker
-from .optimization_passes import get_default_optimization_passes
+from .optimization_passes import get_default_optimization_passes, OptimizationPassStage
 
 if TYPE_CHECKING:
     from angr.knowledge_plugins.cfg import CFGModel
@@ -40,6 +40,8 @@ class Clinic(Analysis):
                  optimization_passes=None,
                  cfg=None,
                  peephole_optimizations: Optional[Iterable[Union[Type['PeepholeOptimizationStmtBase'],Type['PeepholeOptimizationExprBase']]]]=None,
+                 must_struct: Optional[Set[str]]=None,
+                 variable_kb=None,
                  ):
         if not func.normalized:
             raise ValueError("Decompilation must work on normalized function graphs.")
@@ -48,7 +50,7 @@ class Clinic(Analysis):
 
         self.graph = None
         self.arg_list = None
-        self.variable_kb = None
+        self.variable_kb = variable_kb
 
         self._func_graph: Optional[networkx.DiGraph] = None
         self._ail_manager = None
@@ -59,6 +61,7 @@ class Clinic(Analysis):
         self._sp_tracker_track_memory = sp_tracker_track_memory
         self._cfg: Optional['CFGModel'] = cfg
         self.peephole_optimizations = peephole_optimizations
+        self._must_struct = must_struct
 
         # sanity checks
         if not self.kb.functions:
@@ -110,9 +113,11 @@ class Clinic(Analysis):
     def _analyze(self):
 
         # Set up the function graph according to configurations
+        self._update_progress(0., text="Setting up function graph")
         self._set_function_graph()
 
         # Remove alignment blocks
+        self._update_progress(5., text="Removing alignment blocks")
         self._remove_alignment_blocks()
 
         # if the graph is empty, don't continue
@@ -120,30 +125,43 @@ class Clinic(Analysis):
             return
 
         # Make sure calling conventions of all functions have been recovered
+        self._update_progress(10., text="Recovering calling conventions")
         self._recover_calling_conventions()
 
         # initialize the AIL conversion manager
         self._ail_manager = ailment.Manager(arch=self.project.arch)
 
         # Track stack pointers
+        self._update_progress(15., text="Tracking stack pointers")
         spt = self._track_stack_pointers()
 
         # Convert VEX blocks to AIL blocks and then simplify them
 
+        self._update_progress(20., text="Converting VEX to AIL")
         self._convert_all()
 
         ail_graph = self._make_ailgraph()
 
         # Fix "fake" indirect jumps and calls
+        self._update_progress(25., text="Analyzing simple indirect jumps")
         ail_graph = self._replace_single_target_indirect_transitions(ail_graph)
 
         # Make returns
-        ail_graph = self._make_returns(ail_graph)
+        self._update_progress(30., text="Making return sites")
+        if self.function.prototype is None or not isinstance(self.function.prototype.returnty, SimTypeBottom):
+            ail_graph = self._make_returns(ail_graph)
 
         # Simplify blocks
+        self._update_progress(35., text="Simplifying blocks")
         ail_graph = self._simplify_blocks(ail_graph, stack_pointer_tracker=spt)
 
+        # Run simplification passes
+        self._update_progress(40., text="Running simplifications 1")
+        ail_graph = self._run_simplification_passes(ail_graph,
+                                                    stage=OptimizationPassStage.AFTER_SINGLE_BLOCK_SIMPLIFICATION)
+
         # Simplify the entire function for the first time
+        self._update_progress(45., text="Simplifying function 1")
         self._simplify_function(ail_graph, unify_variables=False)
 
         # clear _blocks_by_addr_and_size so no one can use it again
@@ -151,24 +169,28 @@ class Clinic(Analysis):
         self._blocks_by_addr_and_size = None
 
         # Make call-sites
+        self._update_progress(50., text="Making callsites")
         self._make_callsites(ail_graph, stack_pointer_tracker=spt)
 
         # Simplify the entire function for the second time
-        # TODO: Simplify until reaching a fixed point instead of calling _simplify_function() twice.
-        self._simplify_function(ail_graph, unify_variables=True)
+        self._update_progress(55., text="Simplifying function 2")
         self._simplify_function(ail_graph, unify_variables=True)
 
         # Make function arguments
+        self._update_progress(60., text="Making argument list")
         arg_list = self._make_argument_list()
 
         # Recover variables on AIL blocks
+        self._update_progress(65., text="Recovering variables")
         variable_kb = self._recover_and_link_variables(ail_graph, arg_list)
 
         # Make function prototype
+        self._update_progress(70., text="Making function prototype")
         self._make_function_prototype(arg_list, variable_kb)
 
         # Run simplification passes
-        ail_graph = self._run_simplification_passes(ail_graph)
+        self._update_progress(75., text="Running simplifications 2")
+        ail_graph = self._run_simplification_passes(ail_graph, stage=OptimizationPassStage.AFTER_GLOBAL_SIMPLIFICATION)
 
         self.graph = ail_graph
         self.arg_list = arg_list
@@ -324,46 +346,55 @@ class Clinic(Analysis):
         return simp.result_block
 
     @timethis
-    def _simplify_function(self, ail_graph, unify_variables=False):
+    def _simplify_function(self, ail_graph, unify_variables=False, max_iterations: int=4) -> None:
         """
-        Simplify the entire function.
+        Simplify the entire function until it reaches a fixed point.
+        """
+
+        for _ in range(max_iterations):
+            simplified = self._simplify_function_once(ail_graph, unify_variables=unify_variables)
+            if not simplified:
+                break
+
+    @timethis
+    def _simplify_function_once(self, ail_graph, unify_variables=False):
+        """
+        Simplify the entire function once.
 
         :return:    None
         """
 
-        # Computing reaching definitions
-        rd = self.project.analyses.ReachingDefinitions(subject=self.function, func_graph=ail_graph,
-                                                       observe_callback=self._simplify_function_rd_observe_callback)
-
         simp = self.project.analyses.AILSimplifier(
             self.function,
             func_graph=ail_graph,
             remove_dead_memdefs=self._remove_dead_memdefs,
-            reaching_definitions=rd,
             unify_variables=unify_variables,
         )
-
-        def _handler(node):
-            return simp.blocks.get(node, None)
-
-        AILGraphWalker(ail_graph, _handler, replace_nodes=True).walk()
+        # the function graph has been updated at this point
+        return simp.simplified
 
     @timethis
-    def _run_simplification_passes(self, ail_graph):
+    def _run_simplification_passes(self, ail_graph, stage:int=OptimizationPassStage.AFTER_GLOBAL_SIMPLIFICATION):
 
-        blocks_map = defaultdict(set)
+        addr_and_idx_to_blocks: Dict[Tuple[int,Optional[int]],ailment.Block] = { }
+        addr_to_blocks: Dict[int,Set[ailment.Block]] = defaultdict(set)
 
         # update blocks_map to allow node_addr to node lookup
         def _updatedict_handler(node):
-            blocks_map[node.addr].add(node)
+            addr_and_idx_to_blocks[(node.addr, node.idx)] = node
+            addr_to_blocks[node.addr].add(node)
         AILGraphWalker(ail_graph, _updatedict_handler).walk()
 
         # Run each pass
         for pass_ in self._optimization_passes:
 
+            if pass_.STAGE != stage:
+                continue
+
             analysis = getattr(self.project.analyses, pass_.__name__)
 
-            a = analysis(self.function, blocks=blocks_map, graph=ail_graph)
+            a = analysis(self.function, blocks_by_addr=addr_to_blocks, blocks_by_addr_and_idx=addr_and_idx_to_blocks,
+                         graph=ail_graph)
             if a.out_graph:
                 # use the new graph
                 ail_graph = a.out_graph
@@ -415,12 +446,13 @@ class Clinic(Analysis):
         def _handler(block):
             csm = self.project.analyses.AILCallSiteMaker(block, reaching_definitions=rd)
             if csm.result_block:
-                ail_block = csm.result_block
-                simp = self.project.analyses.AILBlockSimplifier(ail_block,
-                                                                stack_pointer_tracker=stack_pointer_tracker,
-                                                                peephole_optimizations=self.peephole_optimizations,
-                                                                )
-                return simp.result_block
+                if csm.result_block != block:
+                    ail_block = csm.result_block
+                    simp = self.project.analyses.AILBlockSimplifier(ail_block,
+                                                                    stack_pointer_tracker=stack_pointer_tracker,
+                                                                    peephole_optimizations=self.peephole_optimizations,
+                                                                    )
+                    return simp.result_block
             return None
 
         AILGraphWalker(ail_graph, _handler, replace_nodes=True).walk()
@@ -447,8 +479,13 @@ class Clinic(Analysis):
                 ret_val = self.function.calling_convention.ret_val
                 if isinstance(ret_val, SimRegArg):
                     reg = self.project.arch.registers[ret_val.reg_name]
-                    new_stmt.ret_exprs.append(ailment.Expr.Register(None, None, reg[0],
-                                                                    reg[1] * self.project.arch.byte_width))
+                    new_stmt.ret_exprs.append(ailment.Expr.Register(
+                        None,
+                        None,
+                        reg[0],
+                        reg[1] * self.project.arch.byte_width,
+                        reg_name=self.project.arch.translate_register_name(reg[0], reg[1])
+                    ))
                 else:
                     l.warning("Unsupported type of return expression %s.",
                               type(self.function.calling_convention.ret_val))
@@ -504,18 +541,13 @@ class Clinic(Analysis):
         else:
             returnty = SimTypeInt()
 
-        self.function.prototype = SimTypeFunction(func_args, returnty)
+        self.function.prototype = SimTypeFunction(func_args, returnty).with_arch(self.project.arch)
 
     @timethis
     def _recover_and_link_variables(self, ail_graph, arg_list):
 
         # variable recovery
-        tmp_kb = KnowledgeBase(self.project)
-        # remove existing variables for this function
-        if tmp_kb.variables.has_function_manager(self.function.addr):
-            l.warning("Removing existing variable recovery result for function %#x.", self.function.addr)
-            del tmp_kb.variables[self.function.addr]
-        # stack pointers have been removed at this point
+        tmp_kb = KnowledgeBase(self.project) if self.variable_kb is None else self.variable_kb
         vr = self.project.analyses.VariableRecoveryFast(self.function,  # pylint:disable=unused-variable
                                                         func_graph=ail_graph, kb=tmp_kb, track_sp=False,
                                                         func_args=arg_list)
@@ -523,14 +555,27 @@ class Clinic(Analysis):
         tmp_kb.variables[self.function.addr].remove_types()
         # TODO: Type inference for global variables
         # run type inference
+        if self._must_struct:
+            must_struct = set()
+            for var, typevar in vr.var_to_typevar.items():
+                if var.ident in self._must_struct:
+                    must_struct.add(typevar)
+        else:
+            must_struct = None
         try:
-            tp = self.project.analyses.Typehoon(vr.type_constraints, kb=tmp_kb, var_mapping=vr.var_to_typevar)
+            tp = self.project.analyses.Typehoon(vr.type_constraints, kb=tmp_kb, var_mapping=vr.var_to_typevar,
+                                                must_struct=must_struct)
             tp.update_variable_types(self.function.addr, vr.var_to_typevar)
         except Exception:  # pylint:disable=broad-except
             l.warning("Typehoon analysis failed. Variables will not have types. Please report to GitHub.",
                       exc_info=True)
 
-        # TODO: The current mapping implementation is kinda hackish...
+        # Unify SSA variables
+        tmp_kb.variables[self.function.addr].unify_variables()
+        tmp_kb.variables[self.function.addr].assign_unified_variable_names(
+            labels=self.kb.labels,
+            reset=self.variable_kb is None
+        )
 
         # Link variables to each statement
         for block in ail_graph.nodes():
@@ -596,8 +641,7 @@ class Clinic(Analysis):
                 self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt,
                                              arg)
         if not is_expr and stmt.ret_expr:
-            self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt,
-                                         stmt.ret_expr)
+            self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt, stmt.ret_expr)
 
     def _link_variables_on_expr(self, variable_manager, global_variables, block, stmt_idx, stmt, expr):
         """
@@ -614,17 +658,29 @@ class Clinic(Analysis):
         if type(expr) is ailment.Expr.Register:
             # find a register variable
             reg_vars = variable_manager.find_variables_by_atom(block.addr, stmt_idx, expr)
-            # TODO: make sure it is the correct register we are looking for
-            if len(reg_vars) == 1:
-                reg_var, offset = next(iter(reg_vars))
+            final_reg_vars = set()
+            if len(reg_vars) > 1:
+                # take phi variables
+                for reg_var in reg_vars:
+                    if variable_manager.is_phi_variable(reg_var[0]):
+                        final_reg_vars.add(reg_var)
+            else:
+                final_reg_vars = reg_vars
+            if len(final_reg_vars) == 1:
+                reg_var, offset = next(iter(final_reg_vars))
                 expr.variable = reg_var
                 expr.variable_offset = offset
 
         elif type(expr) is ailment.Expr.Load:
-            # import ipdb; ipdb.set_trace()
             variables = variable_manager.find_variables_by_atom(block.addr, stmt_idx, expr)
             if len(variables) == 0:
+                # this is a local variable
                 self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt, expr.addr)
+                if 'reference_variable' in expr.addr.tags and expr.addr.reference_variable is not None:
+                    # copy over the variable to this expr since the variable on a constant is supposed to be a
+                    # reference variable.
+                    expr.variable = expr.addr.reference_variable
+                    expr.variable_offset = expr.addr.reference_variable_offset
             else:
                 if len(variables) > 1:
                     l.error("More than one variable are available for atom %s. Consider fixing it using phi nodes.",
@@ -656,6 +712,20 @@ class Clinic(Analysis):
         elif type(expr) is ailment.Expr.Convert:
             self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt, expr.operand)
 
+        elif type(expr) is ailment.Expr.ITE:
+            variables = variable_manager.find_variables_by_atom(block.addr, stmt_idx, expr)
+            if len(variables) == 1:
+                var, offset = next(iter(variables))
+                expr.variable = var
+                expr.variable_offset = offset
+            else:
+                self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt,
+                                             expr.cond)
+                self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt,
+                                             expr.iftrue)
+                self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt,
+                                             expr.iftrue)
+
         elif isinstance(expr, ailment.Expr.BasePointerOffset):
             variables = variable_manager.find_variables_by_atom(block.addr, stmt_idx, expr)
             if len(variables) == 1:
@@ -668,8 +738,8 @@ class Clinic(Analysis):
             variables = global_variables.get_global_variables(expr.value)
             if variables:
                 var = next(iter(variables))
-                expr.variable = var
-                expr.variable_offset = 0
+                expr.tags['reference_variable'] = var
+                expr.tags['reference_variable_offset'] = None
 
         elif isinstance(expr, ailment.Stmt.Call):
             self._link_variables_on_call(variable_manager, global_variables, block, stmt_idx, expr, is_expr=True)
@@ -706,12 +776,5 @@ class Clinic(Analysis):
         op_type = kwargs.pop('op_type')
         return isinstance(stmt, ailment.Stmt.Call) and op_type == OP_BEFORE
 
-    @staticmethod
-    def _simplify_function_rd_observe_callback(ob_type, **kwargs):
-        if ob_type != 'node':
-            return False
-        op_type = kwargs.pop('op_type')
-        return op_type == OP_AFTER
-
 
 register_analysis(Clinic, 'Clinic')
