@@ -1,6 +1,6 @@
 # pylint:disable=unused-import
 from collections import defaultdict
-from typing import List, Tuple, Any
+from typing import List, Tuple, Optional, Iterable, Union, Type, Set, TYPE_CHECKING
 
 from cle import SymbolType
 
@@ -9,14 +9,24 @@ from .. import Analysis, AnalysesHub
 from .condition_processor import ConditionProcessor
 from .decompilation_options import DecompilationOption
 
+if TYPE_CHECKING:
+    from .peephole_optimizations import PeepholeOptimizationStmtBase, PeepholeOptimizationExprBase
+
 
 class Decompiler(Analysis):
-    def __init__(self, func, cfg=None, options=None, optimization_passes=None, sp_tracker_track_memory=True):
+    def __init__(self, func, cfg=None, options=None, optimization_passes=None, sp_tracker_track_memory=True,
+                 peephole_optimizations: Optional[Iterable[Union[Type['PeepholeOptimizationStmtBase'],Type['PeepholeOptimizationExprBase']]]]=None,
+                 vars_must_struct: Optional[Set[str]]=None,
+                 flavor='pseudocode',
+                 ):
         self.func = func
         self._cfg = cfg
         self._options = options
         self._optimization_passes = optimization_passes
         self._sp_tracker_track_memory = sp_tracker_track_memory
+        self._peephole_optimizations = peephole_optimizations
+        self._vars_must_struct = vars_must_struct
+        self._flavor = flavor
 
         self.clinic = None  # mostly for debugging purposes
         self.codegen = None
@@ -28,6 +38,11 @@ class Decompiler(Analysis):
         if self.func.is_simprocedure:
             return
 
+        try:
+            old_codegen = self.kb.structured_code[(self.func.addr, self._flavor)]
+        except KeyError:
+            old_codegen = None
+
         options_by_class = defaultdict(list)
 
         if self._options:
@@ -36,16 +51,22 @@ class Decompiler(Analysis):
 
         # set global variables
         self._set_global_variables()
+        self._update_progress(5., text='Converting to AIL')
 
         # convert function blocks to AIL blocks
         clinic = self.project.analyses.Clinic(self.func,
                                               kb=self.kb,
+                                              variable_kb=old_codegen._variable_kb if old_codegen is not None else None,
                                               optimization_passes=self._optimization_passes,
                                               sp_tracker_track_memory=self._sp_tracker_track_memory,
                                               cfg=self._cfg,
+                                              peephole_optimizations=self._peephole_optimizations,
+                                              must_struct=self._vars_must_struct,
+                                              progress_callback=lambda p, **kwargs: self._update_progress(p*(70-5)/100.+5, **kwargs),
                                               **self.options_to_params(options_by_class['clinic'])
                                               )
         self.clinic = clinic
+        self._update_progress(70., text='Identifying regions')
 
         if clinic.graph is None:
             # the function is empty
@@ -55,17 +76,25 @@ class Decompiler(Analysis):
 
         # recover regions
         ri = self.project.analyses.RegionIdentifier(self.func, graph=clinic.graph, cond_proc=cond_proc, kb=self.kb)
+        self._update_progress(75., text='Structuring code')
 
         # structure it
         rs = self.project.analyses.RecursiveStructurer(ri.region, cond_proc=cond_proc, kb=self.kb)
+        self._update_progress(80., text='Simplifying regions')
 
         # simplify it
         s = self.project.analyses.RegionSimplifier(rs.result, kb=self.kb)
+        self._update_progress(85., text='Generating code')
 
         codegen = self.project.analyses.StructuredCodeGenerator(self.func, s.result, cfg=self._cfg,
+                                                                flavor=self._flavor,
                                                                 func_args=clinic.arg_list,
                                                                 kb=self.kb,
-                                                                variable_kb=clinic.variable_kb)
+                                                                variable_kb=clinic.variable_kb,
+                                                                expr_comments=old_codegen.expr_comments if old_codegen is not None else None,
+                                                                stmt_comments=old_codegen.stmt_comments if old_codegen is not None else None,
+                                                                **self.options_to_params(options_by_class['codegen']))
+        self._update_progress(90., text='Finishing up')
 
         self.codegen = codegen
 
