@@ -1,8 +1,11 @@
 from .coalesce_same_cascading_ifs import CoalesceSameCascadingIfs
 from .constant_derefs import ConstantDereferences
+from .extended_byte_and_mask import ExtendedByteAndMask
 from .remove_empty_if_body import RemoveEmptyIfBody
+from .remove_redundant_ite_branch import RemoveRedundantITEBranches
 from .single_bit_xor import SingleBitXor
 from .a_sub_a_sub_n import ASubASubN
+from .conv_a_sub0_shr_and import ConvASub0ShrAnd
 from .one_sub_bool import OneSubBool
 from .bool_expr_xor_1 import BoolExprXor1
 from .remove_redundant_bitmasks import RemoveRedundantBitmasks
