@@ -1,4 +1,4 @@
-from typing import Dict
+from typing import Dict, Optional
 import logging
 from collections import defaultdict
 
@@ -92,7 +92,7 @@ class Structurer(Analysis):
     """
     def __init__(self, region, parent_map=None, condition_processor=None):
 
-        self._region = region
+        self._region: GraphRegion = region
         self._parent_map = parent_map
 
         self.cond_proc = condition_processor if condition_processor is not None else ConditionProcessor()
@@ -263,7 +263,7 @@ class Structurer(Analysis):
     @staticmethod
     def _refine_loop_while(loop_node):
 
-        if loop_node.sort == 'while' and loop_node.condition is None:
+        if loop_node.sort == 'while' and loop_node.condition is None and loop_node.sequence_node.nodes:
             # it's an endless loop
             first_node = loop_node.sequence_node.nodes[0]
             if type(first_node) is CodeNode:
@@ -281,7 +281,7 @@ class Structurer(Analysis):
     @staticmethod
     def _refine_loop_dowhile(loop_node):
 
-        if loop_node.sort == 'while' and loop_node.condition is None:
+        if loop_node.sort == 'while' and loop_node.condition is None and loop_node.sequence_node.nodes:
             # it's an endless loop
             last_node = loop_node.sequence_node.nodes[-1]
             if type(last_node) is ConditionalBreakNode:
@@ -348,6 +348,7 @@ class Structurer(Analysis):
         # traverse this node and rewrite all jumps that go to the beginning of the loop to continue
         self._rewrite_jumps_to_continues(seq)
 
+        seq = self._remove_redundant_jumps(seq)
         seq = self._remove_conditional_jumps(seq)
         seq = EmptyNodeRemover(seq).result
 
@@ -475,6 +476,7 @@ class Structurer(Analysis):
         self._merge_same_conditioned_nodes(seq)
         self._structure_common_subexpression_conditions(seq)
         self._make_ites(seq)
+        self._replace_complex_reaching_conditions(seq)
         self._make_condition_nodes(seq)
 
         while True:
@@ -975,6 +977,20 @@ class Structurer(Analysis):
         new_node = CodeNode(SequenceNode(nodes=new_nodes), common_subexpr)
         return new_node
 
+    def _replace_complex_reaching_conditions(self, seq: SequenceNode):
+        for i in range(len(seq.nodes)):
+            node = seq.nodes[i]
+
+            if isinstance(node, CodeNode) and \
+                    node.reaching_condition is not None and \
+                    node.reaching_condition.op == "Or" and \
+                    node.node in self.cond_proc.guarding_conditions:
+                guarding_condition = self.cond_proc.guarding_conditions[node.node]
+                # the op of guarding condition is always "Or"
+                if len(guarding_condition.args) < len(node.reaching_condition.args) or \
+                        guarding_condition.depth < node.reaching_condition.depth:
+                    node.reaching_condition = guarding_condition
+
     def _make_condition_nodes(self, seq):
 
         # make all conditionally-reachable nodes ConditionNodes
@@ -982,7 +998,7 @@ class Structurer(Analysis):
             node = seq.nodes[i]
 
             if isinstance(node, CodeNode):
-                if isinstance(node.node, SequenceNode):
+                if isinstance(node.node, SequenceNode) and node.node not in self._new_sequences:
                     self._make_condition_nodes(node.node)
 
                 if node.reaching_condition is not None and not claripy.is_true(node.reaching_condition):
@@ -1033,6 +1049,77 @@ class Structurer(Analysis):
     # Other methods
     #
 
+    @staticmethod
+    def _remove_redundant_jumps(seq):
+        """
+        Remove all redundant jumps.
+
+        :param SequenceNode seq:    The SequenceNode instance to handle.
+        :return:                    A processed SequenceNode.
+        """
+
+        def _handle_Sequence(node: SequenceNode, **kwargs):
+            if len(node.nodes) > 1:
+                for i in range(len(node.nodes) - 1):
+                    this_node = node.nodes[i]
+                    goto_stmt: Optional[ailment.Stmt.Jump] = None
+                    if isinstance(this_node, ailment.Block) and \
+                            this_node.statements and \
+                            isinstance(this_node.statements[-1], ailment.Stmt.Jump):
+                        goto_stmt: ailment.Stmt.Jump = this_node.statements[-1]
+                    elif isinstance(this_node, MultiNode) and \
+                            this_node.nodes and \
+                            isinstance(this_node.nodes[-1], ailment.Block) and \
+                            this_node.nodes[-1].statements and \
+                            isinstance(this_node.nodes[-1].statements[-1], ailment.Stmt.Jump):
+                        this_node = this_node.nodes[-1]
+                        goto_stmt: ailment.Stmt.Jump = this_node.statements[-1]
+
+                    if goto_stmt is not None:
+                        next_node = node.nodes[i + 1]
+                        if isinstance(goto_stmt.target, ailment.Expr.Const) and \
+                                goto_stmt.target.value == next_node.addr:
+                            # this goto is useless
+                            this_node.statements = this_node.statements[:-1]
+
+            return walker._handle_Sequence(node, **kwargs)
+
+        def _handle_MultiNode(node: MultiNode, **kwargs):
+            if len(node.nodes) > 1:
+                for i in range(len(node.nodes) - 1):
+                    this_node = node.nodes[i]
+                    goto_stmt: Optional[ailment.Stmt.Jump] = None
+                    if isinstance(this_node, ailment.Block) and \
+                            this_node.statements and \
+                            isinstance(this_node.statements[-1], ailment.Stmt.Jump):
+                        goto_stmt: ailment.Stmt.Jump = this_node.statements[-1]
+                    elif isinstance(this_node, MultiNode) and \
+                            this_node.nodes and \
+                            isinstance(this_node.nodes[-1], ailment.Block) and \
+                            this_node.nodes[-1].statements and \
+                            isinstance(this_node.nodes[-1].statements[-1], ailment.Stmt.Jump):
+                        goto_stmt: ailment.Stmt.Jump = this_node.nodes[-1].statements[-1]
+                        this_node = this_node.nodes[-1]
+
+                    if goto_stmt is not None:
+                        next_node = node.nodes[i + 1]
+                        if isinstance(goto_stmt.target, ailment.Expr.Const) and \
+                                goto_stmt.target.value == next_node.addr:
+                            # this goto is useless
+                            this_node.statements = this_node.statements[:-1]
+
+            return walker._handle_MultiNode(node, **kwargs)
+
+        handlers = {
+            SequenceNode: _handle_Sequence,
+            MultiNode: _handle_MultiNode,
+        }
+
+        walker = SequenceWalker(handlers=handlers)
+        walker.walk(seq)
+
+        return seq
+
     @staticmethod
     def _remove_conditional_jumps_from_block(block, parent=None, index=0, label=None):  # pylint:disable=unused-argument
         block.statements = [stmt for stmt in block.statements
