@@ -1,4 +1,4 @@
-from typing import Optional, Dict, Set, Iterable, Union, TYPE_CHECKING
+from typing import Optional, Dict, Set, Iterable, Union
 from functools import reduce
 
 import networkx
@@ -145,12 +145,21 @@ class DepGraph:
             )
         ))
 
+        # concretize addresses where possible
+        concrete_known_pred_addresses = []
+        for address in known_predecessor_addresses:
+            if isinstance(address, claripy.ast.Base):
+                if address.concrete:
+                    concrete_known_pred_addresses.append(address._model_concrete.value)
+            else:
+                concrete_known_pred_addresses.append(address)
+
         unknown_concrete_addresses: Set[int] = set()
         for v in values:
             if isinstance(v, claripy.ast.Base) and v.concrete:
                 v = v._model_concrete.value
             if isinstance(v, int):
-                if v not in known_predecessor_addresses:
+                if v not in concrete_known_pred_addresses:
                     unknown_concrete_addresses.add(v)
 
         for address in unknown_concrete_addresses:
