@@ -1,10 +1,12 @@
-from typing import Optional, List
+from typing import Optional, List, Tuple
 import logging
 from collections import defaultdict
 
+import claripy
 import pyvex
 import ailment
 
+from ...storage.memory_mixins.paged_memory.pages.multi_values import MultiValues
 from ...block import Block
 from ...errors import AngrVariableRecoveryError, SimEngineError
 from ...knowledge_plugins import Function
@@ -19,64 +21,6 @@ from .engine_ail import SimEngineVRAIL
 l = logging.getLogger(name=__name__)
 
 
-class ProcessorState:
-
-    __slots__ = ['_arch', 'sp_adjusted', 'sp_adjustment', 'bp_as_base', 'bp']
-
-    def __init__(self, arch):
-        self._arch = arch
-        # whether we have met the initial stack pointer adjustment
-        self.sp_adjusted = None
-        # how many bytes are subtracted from the stack pointer
-        self.sp_adjustment = arch.bytes if arch.call_pushes_ret else 0
-        # whether the base pointer is used as the stack base of the stack frame or not
-        self.bp_as_base = None
-        # content of the base pointer
-        self.bp = None
-
-    def copy(self):
-        s = ProcessorState(self._arch)
-        s.sp_adjusted = self.sp_adjusted
-        s.sp_adjustment = self.sp_adjustment
-        s.bp_as_base = self.bp_as_base
-        s.bp = self.bp
-        return s
-
-    def merge(self, other):
-        if not self == other:
-            l.warning("Inconsistent merge: %s %s ", self, other)
-
-        # FIXME: none of the following logic makes any sense...
-        if other.sp_adjusted is True:
-            self.sp_adjusted = True
-        self.sp_adjustment = max(self.sp_adjustment, other.sp_adjustment)
-        if other.bp_as_base is True:
-            self.bp_as_base = True
-        if self.bp is None:
-            self.bp = other.bp
-        elif other.bp is not None:  # and self.bp is not None
-            if self.bp == other.bp:
-                pass
-            else:
-                if type(self.bp) is int and type(other.bp) is int:
-                    self.bp = max(self.bp, other.bp)
-                else:
-                    self.bp = None
-        return self
-
-    def __eq__(self, other):
-        if not isinstance(other, ProcessorState):
-            return False
-        return (self.sp_adjusted == other.sp_adjusted and
-                self.sp_adjustment == other.sp_adjustment and
-                self.bp == other.bp and
-                self.bp_as_base == other.bp_as_base)
-
-    def __repr__(self):
-        return "<ProcessorState %s%#x%s %s>" % (self.bp, self.sp_adjustment,
-            " adjusted" if self.sp_adjusted else "", self.bp_as_base)
-
-
 class VariableRecoveryFastState(VariableRecoveryStateBase):
     """
     The abstract state of variable recovery analysis.
@@ -84,14 +28,12 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
     :ivar KeyedRegion stack_region: The stack store.
     :ivar KeyedRegion register_region:  The register store.
     """
-    def __init__(self, block_addr, analysis, arch, func, stack_region=None, register_region=None,
-                 typevars=None, type_constraints=None, delayed_type_constraints=None, processor_state=None):
+    def __init__(self, block_addr, analysis, arch, func, stack_region=None, register_region=None, global_region=None,
+                 typevars=None, type_constraints=None, delayed_type_constraints=None, project=None):
 
         super().__init__(block_addr, analysis, arch, func, stack_region=stack_region, register_region=register_region,
-                         typevars=typevars, type_constraints=type_constraints,
-                         delayed_type_constraints=delayed_type_constraints)
-
-        self.processor_state = ProcessorState(self.arch) if processor_state is None else processor_state
+                         global_region=global_region, typevars=typevars, type_constraints=type_constraints,
+                         delayed_type_constraints=delayed_type_constraints, project=project)
 
     def __repr__(self):
         return "<VRAbstractState@%#x: %d register variables, %d stack variables>" % (
@@ -111,15 +53,16 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
             self.function,
             stack_region=self.stack_region.copy(),
             register_region=self.register_region.copy(),
+            global_region=self.global_region.copy(),
             typevars=self.typevars.copy(),
             type_constraints=self.type_constraints.copy(),
-            processor_state=self.processor_state.copy(),
             delayed_type_constraints=self.delayed_type_constraints.copy(),
+            project=self.project,
         )
 
         return state
 
-    def merge(self, other, successor=None):
+    def merge(self, others: Tuple['VariableRecoveryFastState'], successor=None) -> Tuple['VariableRecoveryFastState',bool]:
         """
         Merge two abstract states.
 
@@ -127,26 +70,40 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
         create a phi variable V' for each variable V that is defined in A, and then replace all existence of V with V'
         in the merged abstract state.
 
-        :param VariableRecoveryState other: The other abstract state to merge.
-        :return:                            The merged abstract state.
-        :rtype:                             VariableRecoveryState
+        :param others: Other abstract states to merge.
+        :return:       The merged abstract state.
         """
 
-        replacements = {}
-        if successor in self.dominance_frontiers:
-            replacements = self._make_phi_variables(successor, self, other)
+        self.phi_variables = {}  # A mapping from original variable and its corresponding phi variable
+        self.successor_block_addr = successor
 
-        merged_stack_region = self.stack_region.copy().replace(replacements).merge(other.stack_region,
-                                                                                   replacements=replacements)
-        merged_register_region = self.register_region.copy().replace(replacements).merge(other.register_region,
-                                                                                         replacements=replacements)
-        merged_typevars = self.typevars.merge(other.typevars)
-        merged_typeconstraints = self.type_constraints.copy() | other.type_constraints
+        merged_stack_region = self.stack_region.copy()
+        merged_stack_region.set_state(self)
+        merge_occurred = merged_stack_region.merge([other.stack_region for other in others], None)
+
+        merged_register_region = self.register_region.copy()
+        merged_register_region.set_state(self)
+        merge_occurred |= merged_register_region.merge([other.register_region for other in others], None)
+
+        merged_global_region = self.global_region.copy()
+        merged_global_region.set_state(self)
+        merge_occurred |= merged_global_region.merge([other.global_region for other in others], None)
+
+        merged_typevars = self.typevars
+        merged_typeconstraints = self.type_constraints.copy()
         delayed_typeconstraints = self.delayed_type_constraints.copy()
-        for v, cons in other.delayed_type_constraints.items():
-            delayed_typeconstraints[v] |= cons
+        for other in others:
+            merged_typevars = merged_typevars.merge(other.typevars)
+            merged_typeconstraints |= other.type_constraints
+            for v, cons in other.delayed_type_constraints.items():
+                delayed_typeconstraints[v] |= cons
+
+        merge_occurred |= self.typevars != merged_typevars
+        merge_occurred |= self.type_constraints != merged_typeconstraints
+        merge_occurred |= self.delayed_type_constraints != self.delayed_type_constraints
+
         # add subtype constraints for all replacements
-        for v0, v1 in replacements.items():
+        for v0, v1 in self.phi_variables.items():
             # v0 will be replaced by v1
             if not merged_typevars.has_type_variable_for(v1, None):
                 merged_typevars.add_type_variable(v1, None, TypeVariable())
@@ -164,6 +121,10 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
                                       )
             delayed_typeconstraints[v1].add(equivalence)
 
+        # clean up
+        self.phi_variables = {}
+        self.successor_block_addr = None
+
         state = VariableRecoveryFastState(
             successor,
             self._analysis,
@@ -171,13 +132,14 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
             self.function,
             stack_region=merged_stack_region,
             register_region=merged_register_region,
+            global_region=merged_global_region,
             typevars=merged_typevars,
             type_constraints=merged_typeconstraints,
             delayed_type_constraints=delayed_typeconstraints,
-            processor_state=self.processor_state.copy().merge(other.processor_state),
+            project=self.project,
         )
 
-        return state
+        return state, merge_occurred
 
     #
     # Util methods
@@ -206,7 +168,7 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  #pylint:disa
     accurately. However, it is not a requirement.
     """
 
-    def __init__(self, func, func_graph=None, max_iterations=1, low_priority=False, track_sp=True,
+    def __init__(self, func, func_graph=None, max_iterations=2, low_priority=False, track_sp=True,
                  func_args: Optional[List[SimVariable]]=None):
         """
 
@@ -270,42 +232,51 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  #pylint:disa
     def _pre_job_handling(self, job):
         self._job_ctr += 1
         if self._low_priority:
-            self._release_gil(self._job_ctr, 5, 0.0001)
+            self._release_gil(self._job_ctr, 5, 0.000001)
 
     def _initial_abstract_state(self, node):
-
-        # annotate the stack pointer
-        # concrete_state.regs.sp = concrete_state.regs.sp.annotate(StackLocationAnnotation(8))
-
-        # give it enough stack space
-        # concrete_state.regs.bp = concrete_state.regs.sp + 0x100000
-
-        state = VariableRecoveryFastState(node.addr, self, self.project.arch, self.function,
+        state = VariableRecoveryFastState(node.addr, self, self.project.arch, self.function, project=self.project,
                                           )
+        initial_sp = state.stack_address(self.project.arch.bytes if self.project.arch.call_pushes_ret else 0)
+        state.register_region.store(self.project.arch.sp_offset, initial_sp)
+        # give it enough stack space
+        state.register_region.store(self.project.arch.bp_offset, initial_sp + 0x100000)
+
         # put a return address on the stack if necessary
         if self.project.arch.call_pushes_ret:
             ret_addr_offset = self.project.arch.bytes
             ret_addr_var = SimStackVariable(ret_addr_offset, self.project.arch.bytes, base='bp', name='ret_addr',
                                             region=self.function.addr, category='return_address',
                                             )
-            state.stack_region.add_variable(ret_addr_offset, ret_addr_var)
+            ret_addr = claripy.BVS("ret_addr", self.project.arch.bits)
+            ret_addr = state.annotate_with_variables(ret_addr, [(0, ret_addr_var)])
+            state.stack_region.store(state.stack_addr_from_offset(ret_addr_offset),
+                                     ret_addr,
+                                     endness=self.project.arch.memory_endness)
 
         if self._func_args:
             for arg in self._func_args:
                 if isinstance(arg, SimRegisterVariable):
-                    state.register_region.set_variable(arg.reg, arg)
+                    v = claripy.BVS("reg_arg", arg.bits)
+                    v = state.annotate_with_variables(v, [(0, arg)])
+                    state.register_region.store(arg.reg, v)
                     self.variable_manager[self.function.addr].add_variable('register', arg.reg, arg)
                 elif isinstance(arg, SimStackVariable):
-                    state.stack_region.set_variable(arg.offset, arg)
+                    v = claripy.BVS("stack_arg", arg.bits)
+                    v = state.annotate_with_variables(v, [(0, arg)])
+                    state.stack_region.store(state.stack_addr_from_offset(arg.offset),
+                                             v,
+                                             endness=self.project.arch.memory_endness,
+                                             )
                     self.variable_manager[self.function.addr].add_variable('stack', arg.offset, arg)
                 else:
                     raise TypeError("Unsupported function argument type %s." % type(arg))
 
         return state
 
-    def _merge_states(self, node, *states):
-
-        return states[0].merge(states[1], successor=node.addr)
+    def _merge_states(self, node, *states: VariableRecoveryFastState):
+        merged_state, merge_occurred = states[0].merge(states[1:], successor=node.addr)
+        return merged_state, not merge_occurred
 
     def _run_on_node(self, node, state):
         """
@@ -325,14 +296,14 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  #pylint:disa
             # VEX mode, get the block again
             block = self.project.factory.block(node.addr, node.size, opt_level=1, cross_insn_opt=False)
 
-        if node.addr in self._instates:
-            prev_state = self._instates[node.addr]
-            if input_state == prev_state:
-                l.debug('Skip node %#x as we have reached a fixed-point', node.addr)
-                return False, input_state
-            else:
-                l.debug('Merging input state of node %#x with the previous state.', node.addr)
-                input_state = prev_state.merge(input_state, successor=node.addr)
+        # if node.addr in self._instates:
+        #     prev_state: VariableRecoveryFastState = self._instates[node.addr]
+        #     if input_state == prev_state:
+        #         l.debug('Skip node %#x as we have reached a fixed-point', node.addr)
+        #         return False, input_state
+        #     else:
+        #         l.debug('Merging input state of node %#x with the previous state.', node.addr)
+        #         input_state, _ = prev_state.merge((input_state,), successor=node.addr)
 
         state = input_state.copy()
         state.block_addr = node.addr
@@ -356,6 +327,7 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  #pylint:disa
         pass
 
     def _post_analysis(self):
+        self.variable_manager['global'].assign_variable_names(labels=self.kb.labels)
         self.variable_manager[self.function.addr].assign_variable_names()
 
         for addr, state in self._outstates.items():
@@ -441,23 +413,28 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  #pylint:disa
 
         if self._track_sp and block.addr in self._node_to_cc:
             # readjusting sp at the end for blocks that end in a call
+            sp: MultiValues = state.register_region.load(self.project.arch.sp_offset, size=self.project.arch.bytes)
+            sp_v = sp.one_value()
+            if sp_v is None:
+                l.warning("Unexpected stack pointer value at the end of the function. Pick the first one.")
+                sp_v = next(iter(next(iter(sp.values.values()))))
+
             adjusted = False
-            if state.processor_state.sp_adjustment is None:
-                state.processor_state.sp_adjustment = 0
             cc = self._node_to_cc[block.addr]
             if cc is not None and cc.sp_delta is not None:
-                state.processor_state.sp_adjustment += cc.sp_delta
-                state.processor_state.sp_adjusted = True
+                sp_v += cc.sp_delta
                 adjusted = True
-                l.debug('Adjusting stack pointer at end of block %#x with offset %+#x.',
-                        block.addr, state.processor_state.sp_adjustment)
+                l.debug('Adjusting stack pointer at end of block %#x with offset %+#x.', block.addr, cc.sp_delta)
 
             if not adjusted:
                 # make a guess
                 # of course, this will fail miserably if the function called is not cdecl
                 if self.project.arch.call_pushes_ret:
-                    state.processor_state.sp_adjustment += self.project.arch.bytes
-                    state.processor_state.sp_adjusted = True
+                    sp_v += self.project.arch.bytes
+                    adjusted = True
+
+            if adjusted:
+                state.register_region.store(self.project.arch.sp_offset, sp_v)
 
 
 from angr.analyses import AnalysesHub
