@@ -12,6 +12,7 @@ from archinfo.arch_soot import SootAddressDescriptor, ArchSoot
 import cle
 
 from .misc.ux import deprecated
+from .errors import AngrNoPluginError
 
 l = logging.getLogger(name=__name__)
 
@@ -117,6 +118,9 @@ class Project:
                 l.warning("You provided CLE options to angr but you also provided a completed cle.Loader object!")
             self.loader = thing
             self.filename = self.loader.main_object.binary
+        elif isinstance(thing, cle.Backend):
+            self.filename = thing.binary
+            self.loader = cle.Loader(thing, **load_options)
         elif hasattr(thing, 'read') and hasattr(thing, 'seek'):
             l.info("Loading binary from stream")
             self.filename = None
@@ -188,8 +192,9 @@ class Project:
         self.factory = AngrObjectFactory(self, default_engine=engine)
 
         # Step 4.2: Analyses
-        self.analyses = AnalysesHub(self)
-        self.analyses.use_plugin_preset(analyses_preset if analyses_preset is not None else 'default')
+        self._analyses_preset = analyses_preset
+        self.analyses = None
+        self._initialize_analyses_hub()
 
         # Step 4.3: ...etc
         self.kb = KnowledgeBase(self, name="global")
@@ -221,6 +226,13 @@ class Project:
         # Step 7: Run OS-specific configuration
         self.simos.configure_project()
 
+    def _initialize_analyses_hub(self):
+        """
+        Initializes self.analyses using a given preset.
+        """
+        self.analyses = AnalysesHub(self)
+        self.analyses.use_plugin_preset(self._analyses_preset if self._analyses_preset is not None else 'default')
+
     def _register_object(self, obj, sim_proc_arch):
         """
         This scans through an objects imports and hooks them with simprocedures from our library whenever possible
@@ -331,7 +343,16 @@ class Project:
                 else:
                     if not func.is_weak:
                         l.info("Using stub SimProcedure for unresolved %s", export.name)
-                        self.hook_symbol(export.rebased_addr, missing_libs[0].get(export.name, sim_proc_arch))
+                        the_lib = missing_libs[0]
+                        if export.name and export.name.startswith("_Z"):
+                            # GNU C++ name. Use a C++ library to create the stub
+                            if 'libstdc++.so' in SIM_LIBRARIES:
+                                the_lib = SIM_LIBRARIES['libstdc++.so']
+                            else:
+                                l.critical("Does not find any C++ library in SIM_LIBRARIES. We may not correctly "
+                                           "create the stub or resolve the function prototype for name %s.", export.name)
+
+                        self.hook_symbol(export.rebased_addr, the_lib.get(export.name, sim_proc_arch))
 
             # Step 2.5: If 2.4 didn't work (we have NO SimLibraries to work with), just
             # use the vanilla ReturnUnconstrained, assuming that this isn't a weak func
@@ -640,12 +661,21 @@ class Project:
         try:
             store_func, load_func = self.store_function, self.load_function
             self.store_function, self.load_function = None, None
-            return dict(self.__dict__)
+            # ignore analyses. we re-initialize analyses when restoring from pickling so that we do not lose any newly
+            # added analyses classes
+            d = dict((k, v) for k, v in self.__dict__.items() if k not in {'analyses', })
+            return d
         finally:
             self.store_function, self.load_function = store_func, load_func
 
     def __setstate__(self, s):
         self.__dict__.update(s)
+        try:
+            self._initialize_analyses_hub()
+        except AngrNoPluginError:
+            l.warning("Plugin preset %s does not exist any more. Fall back to the default preset.")
+            self._analyses_preset = 'default'
+            self._initialize_analyses_hub()
 
     def _store(self, container):
         # If container is a filename.
