@@ -51,6 +51,7 @@ State::State(uc_engine *_uc, uint64_t cache_key):uc(_uc) {
 	arch = *((uc_arch*)uc); // unicorn hides all its internals...
 	mode = *((uc_mode*)((uc_arch*)uc + 1));
 	curr_block_details.reset();
+	symbolic_read_in_progress = false;
 }
 
 /*
@@ -111,15 +112,6 @@ uc_err State::start(address_t pc, uint64_t step) {
 		return UC_ERR_MAP;
 	}
 
-	// Initialize slice tracker for registers and flags
-	for (auto &reg_entry: vex_to_unicorn_map) {
-		reg_instr_slice.emplace(reg_entry.first, std::vector<instr_details_t>());
-	}
-
-	for (auto &cpu_flag_entry: cpu_flags) {
-		reg_instr_slice.emplace(cpu_flag_entry.first, std::vector<instr_details_t>());
-	}
-
 	uc_err out = uc_emu_start(uc, unicorn_next_instr_addr, 0, 0, 0);
 	if (out == UC_ERR_OK && stop_details.stop_reason == STOP_NOSTART && get_instruction_pointer() == 0) {
 		// handle edge case where we stop because we reached our bogus stop address (0)
@@ -252,7 +244,7 @@ void State::commit() {
 	}
 	if (curr_block_details.symbolic_instrs.size() > 0) {
 		for (auto &symbolic_instr: curr_block_details.symbolic_instrs) {
-			compute_slice_of_instrs_for_vex_temps(symbolic_instr);
+			compute_slice_of_instr(symbolic_instr);
 			// Save all concrete memory dependencies of the block
 			save_concrete_memory_deps(symbolic_instr);
 		}
@@ -261,10 +253,14 @@ void State::commit() {
 	// Clear all block level taint status trackers and symbolic instruction list
 	block_symbolic_registers.clear();
 	block_concrete_registers.clear();
+	block_instr_concrete_regs.clear();
 	curr_block_details.reset();
-	instr_slice_details_map.clear();
-	mem_reads_map.clear();
+	block_mem_reads_data.clear();
+	block_mem_reads_map.clear();
 	mem_writes_taint_map.clear();
+	taint_engine_next_instr_address = 0;
+	taint_engine_stop_mem_read_instruction = 0;
+	taint_engine_stop_mem_read_size = 0;
 	return;
 }
 
@@ -704,42 +700,44 @@ void State::handle_write(address_t address, int size, bool is_interrupt) {
 	mem_writes.push_back(record);
 }
 
-void State::compute_slice_of_instrs(address_t instr_addr, const instruction_taint_entry_t &instr_taint_entry) {
-	instr_slice_details_t instr_slice_details;
-	for (auto &dependency: instr_taint_entry.dependencies.at(TAINT_ENTITY_REG)) {
-		auto dep_reg_slice_entry = reg_instr_slice.find(dependency.reg_offset);
-		if (dep_reg_slice_entry == reg_instr_slice.end()) {
-			// We don't care about slice of this register because it's artificial, blacklisted or something else.
-			continue;
-		}
-		if (!is_symbolic_register(dependency.reg_offset, dependency.value_size)) {
-			auto dep_reg_slice_instrs = dep_reg_slice_entry->second;
-			if (dep_reg_slice_instrs.size() == 0) {
-				// The register was not modified in this block by any preceding instruction
-				// and so it's value at start of block is a dependency of the block
-				instr_slice_details.concrete_registers.emplace(dependency.reg_offset, dependency.value_size);
-			}
-			else {
-				// The register was modified by some instructions in the block. We add those
-				// instructions to the slice of this instruction and also any instructions
-				// they depend on
-				for (auto &dep_reg_slice_instr: dep_reg_slice_instrs) {
-					auto& dep_instr_slice_details = instr_slice_details_map.at(dep_reg_slice_instr.instr_addr);
-					instr_slice_details.concrete_registers.insert(dep_instr_slice_details.concrete_registers.begin(), dep_instr_slice_details.concrete_registers.end());
-					instr_slice_details.dependent_instrs.insert(dep_instr_slice_details.dependent_instrs.begin(), dep_instr_slice_details.dependent_instrs.end());
-					instr_slice_details.dependent_instrs.emplace(dep_reg_slice_instr);
-				}
-			}
-		}
-	}
-	instr_slice_details_map.emplace(instr_addr, instr_slice_details);
-	return;
-}
-
-void State::compute_slice_of_instrs_for_vex_temps(instr_details_t &instr) {
-	std::queue<std::pair<taint_entity_t, address_t>> temps_to_process;
+void State::compute_slice_of_instr(instr_details_t &instr) {
+	// Compute block slice of instruction needed to setup concrete registers needed by it and also save values of
+	// registers not changed from start of the block
+	std::unordered_set<address_t> instrs_to_process;
+	bool all_dep_regs_concrete = false;
 	auto &block_taint_entry = block_taint_cache.at(curr_block_details.block_addr);
 	auto &instr_taint_entry = block_taint_entry.block_instrs_taint_data_map.at(instr.instr_addr);
+
+	// Save values of registers not modified from start of block till instruction
+	auto instr_concrete_regs_it = block_instr_concrete_regs.find(instr.instr_addr);
+	if (instr_concrete_regs_it == block_instr_concrete_regs.end()) {
+		// No entry for current instruction in list of concrete regs => all registers were concrete
+		// An entry was not added because no symbolic taint was present to propagate
+		all_dep_regs_concrete = true;
+	}
+	for (auto &dependency: instr_taint_entry.unmodified_dep_regs) {
+		if (!is_valid_dependency_register(dependency.first) ||
+		    (!all_dep_regs_concrete && instr_concrete_regs_it->second.count(dependency.first) == 0)) {
+			// Register is not a valid dependency or was not concrete before instruction was executed. Do not save value.
+			continue;
+		}
+		auto reg_offset = dependency.first;
+		auto reg_size = dependency.second;
+		auto reg_val = block_start_reg_values.at(reg_offset);
+		reg_val.size = reg_size;
+		instr.reg_deps.insert(reg_val);
+	}
+
+	// List of instructions modifying a register dependency. Their slice needs to be computed.
+	for (auto &dep_reg_modifier: instr_taint_entry.dep_reg_modifier_addr) {
+		if (!all_dep_regs_concrete && instr_concrete_regs_it->second.count(dep_reg_modifier.first) == 0) {
+			// Register was not concrete before instruction was executed. Do not compute slice.
+			continue;
+		}
+		instrs_to_process.emplace(dep_reg_modifier.second);
+	}
+
+	// List of instructions modifying a VEX temp dependency. Their slice needs to be computed.
 	for (auto &dependency: instr_taint_entry.dependencies.at(TAINT_ENTITY_TMP)) {
 		auto vex_temp_deps_entry = block_taint_entry.vex_temp_deps.find(dependency);
 		if (vex_temp_deps_entry == block_taint_entry.vex_temp_deps.end()) {
@@ -747,53 +745,30 @@ void State::compute_slice_of_instrs_for_vex_temps(instr_details_t &instr) {
 			continue;
 		}
 		address_t vex_setter_instr = vex_temp_deps_entry->second.first;
-		if (!is_symbolic_temp(dependency.tmp_id)) {
-			temps_to_process.emplace(std::make_pair(dependency, vex_setter_instr));
+		if ((vex_setter_instr != instr.instr_addr) && !is_symbolic_temp(dependency.tmp_id)) {
+			instrs_to_process.emplace(vex_setter_instr);
 		}
 	}
-	std::vector<instr_details_t> new_dep_instrs(instr.instr_deps.begin(), instr.instr_deps.end());
-	instr.instr_deps.clear();
-	while (!temps_to_process.empty()) {
-		auto &vex_tmp = temps_to_process.front();
-		if (vex_tmp.second != instr.instr_addr) {
-			auto &tmp_instr_slice = instr_slice_details_map.at(vex_tmp.second);
-			auto &tmp_instr_details = block_taint_entry.block_instrs_taint_data_map.at(vex_tmp.second);
-			for (auto &reg: tmp_instr_slice.concrete_registers) {
-				auto reg_offset = reg.first;
-				auto reg_val = block_start_reg_values.at(reg_offset);
-				reg_val.size = reg.second;
-				instr.reg_deps.insert(reg_val);
-			}
-			new_dep_instrs.insert(new_dep_instrs.end(), tmp_instr_slice.dependent_instrs.begin(), tmp_instr_slice.dependent_instrs.end());
-			new_dep_instrs.emplace_back(compute_instr_details(vex_tmp.second, tmp_instr_details));
-		}
-		temps_to_process.pop();
-		for (auto &dep_vex_tmp: block_taint_entry.vex_temp_deps.at(vex_tmp.first).second) {
-			auto vex_temp_deps_entry = block_taint_entry.vex_temp_deps.find(dep_vex_tmp);
-			if (vex_temp_deps_entry == block_taint_entry.vex_temp_deps.end()) {
-				// No dependency entries for this VEX temp
-				continue;
-			}
-			address_t vex_setter_instr = vex_temp_deps_entry->second.first;
-			temps_to_process.push(std::make_pair(dep_vex_tmp, vex_setter_instr));
-		}
-	}
-	for (auto &dep_instr: new_dep_instrs) {
-		compute_slice_of_instrs_for_vex_temps(dep_instr);
-		instr.reg_deps.insert(dep_instr.reg_deps.begin(), dep_instr.reg_deps.end());
-		instr.instr_deps.insert(dep_instr.instr_deps.begin(), dep_instr.instr_deps.end());
-		dep_instr.reg_deps.clear();
-		dep_instr.instr_deps.clear();
-		instr.instr_deps.insert(dep_instr);
+	for (auto &instr_to_process_addr: instrs_to_process) {
+		auto &instr_to_process_taint_entry = block_taint_entry.block_instrs_taint_data_map.at(instr_to_process_addr);
+		instr_details_t instr_details = compute_instr_details(instr_to_process_addr, instr_to_process_taint_entry);
+		compute_slice_of_instr(instr_details);
+		instr.reg_deps.insert(instr_details.reg_deps.begin(), instr_details.reg_deps.end());
+		instr.instr_deps.insert(instr_details.instr_deps.begin(), instr_details.instr_deps.end());
+		instr_details.reg_deps.clear();
+		instr_details.instr_deps.clear();
+		instr.instr_deps.insert(instr_details);
 	}
 	return;
 }
 
-block_taint_entry_t State::process_vex_block(IRSB *vex_block, address_t address) {
+void State::process_vex_block(IRSB *vex_block, address_t address) {
 	block_taint_entry_t block_taint_entry;
 	instruction_taint_entry_t instruction_taint_entry;
 	bool started_processing_instructions;
 	address_t curr_instr_addr;
+	std::unordered_map<vex_reg_offset_t, address_t> last_reg_modifier_instr;
+	std::unordered_set<vex_reg_offset_t> modified_regs;
 
 	started_processing_instructions = false;
 	block_taint_entry.has_unsupported_stmt_or_expr_type = false;
@@ -804,13 +779,10 @@ block_taint_entry_t State::process_vex_block(IRSB *vex_block, address_t address)
 			{
 				taint_entity_t sink;
 				std::unordered_set<taint_entity_t> srcs;
-				std::pair<vex_reg_offset_t, std::pair<int64_t, bool>> modified_reg_data;
 
 				sink.entity_type = TAINT_ENTITY_REG;
 				sink.instr_addr = curr_instr_addr;
 				sink.reg_offset = stmt->Ist.Put.offset;
-				modified_reg_data.first = stmt->Ist.Put.offset;
-				modified_reg_data.second.second = false;
 				auto result = process_vex_expr(stmt->Ist.Put.data, vex_block->tyenv, curr_instr_addr, false);
 				if (result.has_unsupported_expr) {
 					block_taint_entry.has_unsupported_stmt_or_expr_type = true;
@@ -818,7 +790,6 @@ block_taint_entry_t State::process_vex_block(IRSB *vex_block, address_t address)
 					break;
 				}
 				sink.value_size = result.value_size;
-				modified_reg_data.second.first = sink.value_size;
 				// Flatten list of taint sources and also save them as dependencies of instruction
 				// TODO: Should we not save dependencies if sink is an artificial register?
 				for (auto &entry: result.taint_sources) {
@@ -828,18 +799,13 @@ block_taint_entry_t State::process_vex_block(IRSB *vex_block, address_t address)
 				instruction_taint_entry.taint_sink_src_map.emplace_back(sink, srcs);
 				instruction_taint_entry.mem_read_size += result.mem_read_size;
 				instruction_taint_entry.has_memory_read |= (result.mem_read_size != 0);
-				// Check if sink is also source of taint
-				if (result.taint_sources.at(sink.entity_type).count(sink)) {
-					modified_reg_data.second.second = true;
-				}
 				// Store ITE condition entities. Also, store them as dependencies of instruction.
 				for (auto &entry: result.ite_cond_entities) {
 					instruction_taint_entry.ite_cond_entity_list.insert(entry.second.begin(), entry.second.end());
 					instruction_taint_entry.dependencies.at(entry.first).insert(entry.second.begin(), entry.second.end());
 				}
-				if ((modified_reg_data.first != arch_pc_reg_vex_offset()) && reg_instr_slice.count(modified_reg_data.first) != 0) {
-					instruction_taint_entry.modified_regs.emplace_back(modified_reg_data);
-				}
+				// Mark this register as modified by this instruction for updating register setter later
+				modified_regs.emplace(sink.reg_offset);
 				break;
 			}
 			case Ist_WrTmp:
@@ -956,6 +922,26 @@ block_taint_entry_t State::process_vex_block(IRSB *vex_block, address_t address)
 			{
 				// Save dependencies of previous instruction and clear it
 				if (started_processing_instructions) {
+					for (auto &dep: instruction_taint_entry.dependencies.at(TAINT_ENTITY_REG)) {
+						auto entry = last_reg_modifier_instr.find(dep.reg_offset);
+						if (entry != last_reg_modifier_instr.end()) {
+							instruction_taint_entry.dep_reg_modifier_addr.emplace(dep.reg_offset, entry->second);
+						}
+						else {
+							instruction_taint_entry.unmodified_dep_regs.emplace(dep.reg_offset, dep.value_size);
+						}
+					}
+					// Update last modified instruction address for registers modified by this instruction
+					for (auto &modified_reg: modified_regs) {
+						auto last_modifier_entry = last_reg_modifier_instr.find(modified_reg);
+						if (last_modifier_entry == last_reg_modifier_instr.end()) {
+							last_reg_modifier_instr.emplace(modified_reg, curr_instr_addr);
+						}
+						else {
+							last_modifier_entry->second = curr_instr_addr;
+						}
+					}
+					modified_regs.clear();
 					// TODO: Many instructions will not have dependencies. Can we save memory by not storing info for them?
 					block_taint_entry.block_instrs_taint_data_map.emplace(curr_instr_addr, instruction_taint_entry);
 				}
@@ -1036,7 +1022,8 @@ block_taint_entry_t State::process_vex_block(IRSB *vex_block, address_t address)
 	}
 	// Save last instruction's entry
 	block_taint_entry.block_instrs_taint_data_map.emplace(curr_instr_addr, instruction_taint_entry);
-	return block_taint_entry;
+	block_taint_cache.emplace(address, block_taint_entry);
+	return;
 }
 
 std::set<instr_details_t> State::get_list_of_dep_instrs(const instr_details_t &instr) const {
@@ -1412,7 +1399,7 @@ taint_status_result_t State::get_final_taint_status(const std::unordered_set<tai
 				// Address is concrete so we check result of the memory read
 				mem_read_result_t mem_read_result;
 				try {
-					mem_read_result = mem_reads_map.at(taint_source.instr_addr);
+					mem_read_result = block_mem_reads_map.at(taint_source.instr_addr);
 				}
 				catch (std::out_of_range const&) {
 					assert(false && "[sim_unicorn] Taint sink depends on a read not executed yet! This should not happen!");
@@ -1442,6 +1429,21 @@ int32_t State::get_vex_expr_result_size(IRExpr *expr, IRTypeEnv* tyenv) const {
 	return sizeofIRType(expr_type);
 }
 
+VEXLiftResult* State::lift_block(address_t block_address, int32_t block_size) {
+	VexRegisterUpdates pxControl = VexRegUpdUnwindregsAtMemAccess;
+	std::unique_ptr<uint8_t[]> instructions(new uint8_t[block_size]);
+	address_t lift_address;
+
+	if ((arch == UC_ARCH_ARM) && is_thumb_mode()) {
+		lift_address = block_address | 1;
+	}
+	else {
+		lift_address = block_address;
+	}
+	uc_mem_read(this->uc, lift_address, instructions.get(), block_size);
+	return vex_lift(vex_guest, vex_archinfo, instructions.get(), lift_address, 99, block_size, 1, 0, 1, 1, 0, pxControl);
+}
+
 void State::mark_register_symbolic(vex_reg_offset_t reg_offset, int64_t reg_size) {
 	// Mark register as symbolic in the state in current block
 	if (is_blacklisted_register(reg_offset)) {
@@ -1555,6 +1557,15 @@ void State::propagate_taints() {
 	for (; instr_taint_data_entries_it != instr_taint_data_stop_it && !stopped; ++instr_taint_data_entries_it) {
 		address_t curr_instr_addr = instr_taint_data_entries_it->first;
 		auto& curr_instr_taint_entry = instr_taint_data_entries_it->second;
+		std::unordered_map<vex_reg_offset_t, int64_t> concrete_reg_deps;
+
+		// Save list of register dependencies of current instruction which are concrete for slice computation later
+		for (auto &reg_dep: curr_instr_taint_entry.dependencies.at(TAINT_ENTITY_REG)) {
+			if (!is_symbolic_register(reg_dep.reg_offset, reg_dep.value_size)) {
+				concrete_reg_deps.emplace(std::make_pair(reg_dep.reg_offset, reg_dep.value_size));
+			}
+		}
+		block_instr_concrete_regs.emplace(curr_instr_addr, concrete_reg_deps);
 		if (curr_instr_taint_entry.has_memory_read) {
 			// Pause taint propagation to process the memory read and continue from instruction
 			// after the memory read.
@@ -1569,13 +1580,9 @@ void State::propagate_taints() {
 			if (curr_instr_taint_entry.has_memory_write) {
 				mem_writes_taint_map.emplace(curr_instr_addr, false);
 			}
-			compute_slice_of_instrs(curr_instr_addr, curr_instr_taint_entry);
-			update_register_slice(curr_instr_addr, curr_instr_taint_entry);
 			continue;
 		}
-		compute_slice_of_instrs(curr_instr_addr, curr_instr_taint_entry);
 		propagate_taint_of_one_instr(curr_instr_addr, curr_instr_taint_entry);
-		update_register_slice(curr_instr_addr, curr_instr_taint_entry);
 	}
 	// If we reached here, execution has reached the end of the block
 	if (!stopped) {
@@ -1595,14 +1602,146 @@ void State::propagate_taints() {
 	return;
 }
 
-void State::propagate_taint_of_mem_read_instr_and_continue(const address_t instr_addr) {
+void State::propagate_taint_of_mem_read_instr_and_continue(address_t read_address, int read_size) {
+	memory_value_t memory_read_value;
+	address_t curr_instr_addr;
+
+	auto tainted = find_tainted(read_address, read_size);
 	if (is_symbolic_tracking_disabled()) {
+		if (tainted != -1) {
+			// Symbolic register tracking is disabled but memory location has symbolic data.
+			// We switch to VEX engine then.
+			stop(STOP_SYMBOLIC_READ_SYMBOLIC_TRACKING_DISABLED);
+			return;
+		}
 		// We're not checking symbolic registers so no need to propagate taints
 		return;
 	}
-	auto mem_read_result = mem_reads_map.at(instr_addr);
+
+	// Save info about the memory read
+	memory_read_value.reset();
+	memory_read_value.address = read_address;
+	memory_read_value.size = read_size;
+	if (tainted != -1) {
+		memory_read_value.is_value_symbolic = true;
+	}
+	else {
+		memory_read_value.is_value_symbolic = false;
+		read_memory_value(read_address, read_size, memory_read_value.value, MAX_MEM_ACCESS_SIZE);
+	}
+
+	if (!memory_read_value.is_value_symbolic && !symbolic_read_in_progress && (symbolic_registers.size() == 0) &&
+	    (block_symbolic_registers.size() == 0) && (block_symbolic_temps.size() == 0)) {
+		// The value read from memory is concrete and there are no symbolic registers or VEX temps. No need to propagate
+		// taint. Since we cannot rely on the unicorn engine to find out current instruction correctly, we simply save
+		// the memory read value in a list for now and rebuild the map later if needed using instruction info from VEX
+		// block
+		block_mem_reads_data.emplace_back(memory_read_value);
+		return;
+	}
+
+	if (block_taint_cache.find(curr_block_details.block_addr) == block_taint_cache.end()) {
+		// The VEX statements of current block has not been processed yet. This means symbolic taint is being introduced
+		// by this memory read. Let's process the block, rebuild its memory reads map and find the current instruction
+		// address
+		curr_block_details.vex_lift_result = lift_block(curr_block_details.block_addr, curr_block_details.block_size);
+		if ((curr_block_details.vex_lift_result == NULL) || (curr_block_details.vex_lift_result->size == 0)) {
+			// Failed to lift block to VEX.
+			if (memory_read_value.is_value_symbolic) {
+				// Since we are processing VEX block for the first time, there are no symbolic registers/VEX temps.
+				// Thus, it is sufficient to check if the value read from memory is symbolic.
+				stop(STOP_VEX_LIFT_FAILED);
+			}
+			else {
+				// There are no symbolic registers so let's attempt to execute the block.
+				curr_block_details.vex_lift_failed = true;
+			}
+			return;
+		}
+		process_vex_block(curr_block_details.vex_lift_result->irsb, curr_block_details.block_addr);
+	}
+	auto& block_taint_entry = block_taint_cache.at(curr_block_details.block_addr);
+	if (taint_engine_stop_mem_read_instruction != 0) {
+		// Taint has been propagated and so we can rely on information from taint engine to find current instruction
+		// address and hence update the block's memory reads map
+		curr_instr_addr = taint_engine_stop_mem_read_instruction;
+	}
+	else {
+		// Symbolic taint is being introduced by this memory read so we cannot rely on taint engine to find current
+		// instruction address
+		std::map<address_t, instruction_taint_entry_t>::iterator instr_entry_it = block_taint_entry.block_instrs_taint_data_map.begin();
+		if (block_mem_reads_data.size() > 0) {
+			// There are previous reads that need to be insert into block's memory reads map
+			while (instr_entry_it != block_taint_entry.block_instrs_taint_data_map.end()) {
+				if (instr_entry_it->second.has_memory_read) {
+					mem_read_result_t mem_read_result;
+					while (block_mem_reads_data.size() != 0) {
+						auto &next_mem_read = block_mem_reads_data.front();
+						mem_read_result.memory_values.emplace_back(next_mem_read);
+						mem_read_result.is_mem_read_symbolic |= next_mem_read.is_value_symbolic;
+						mem_read_result.read_size += next_mem_read.size;
+						block_mem_reads_data.erase(block_mem_reads_data.begin());
+						if (mem_read_result.read_size == instr_entry_it->second.mem_read_size) {
+							block_mem_reads_map.emplace(instr_entry_it->first, mem_read_result);
+							break;
+						}
+					}
+					if (block_mem_reads_data.size() == 0) {
+						// All pending reads have been processed and inserted into the map
+						if (block_mem_reads_map.at(instr_entry_it->first).read_size == instr_entry_it->second.mem_read_size) {
+							// Update iterator since all reads for current instruction have been processed. We should
+							// start searching for next instruction with memory read from successor of this instruction.
+							instr_entry_it++;
+						}
+						else {
+							// There are still more reads pending for this instruction. Insert partial result value into the
+							// block's memory reads map
+							block_mem_reads_map.emplace(instr_entry_it->first, mem_read_result);
+						}
+						break;
+					}
+				}
+				instr_entry_it++;
+			}
+			if ((block_mem_reads_data.size() != 0) && (instr_entry_it == block_taint_entry.block_instrs_taint_data_map.end())) {
+				// There are still some pending reads but all instructions in the block have been processed. Something
+				// is wrong.
+				assert(false && "There are pending memory reads to process but full block has been processed. This should not happen!");
+			}
+		}
+		// Find next instruction with memory read
+		while (!instr_entry_it->second.has_memory_read) {
+			instr_entry_it++;
+			if (instr_entry_it == block_taint_entry.block_instrs_taint_data_map.end()) {
+				// Current read does not belong to any possible instruction in current block. This should not happen!
+				assert(false && "Unable to identify instruction for current memory read. This should not happen!");
+			}
+		}
+		curr_instr_addr = instr_entry_it->first;
+		taint_engine_stop_mem_read_instruction = curr_instr_addr;
+		taint_engine_stop_mem_read_size = instr_entry_it->second.mem_read_size;
+		taint_engine_next_instr_address = std::next(instr_entry_it)->first;
+	}
+	auto mem_reads_map_entry = block_mem_reads_map.find(curr_instr_addr);
+	if (mem_reads_map_entry == block_mem_reads_map.end()) {
+		mem_read_result_t mem_read_result;
+		mem_read_result.memory_values.emplace_back(memory_read_value);
+		mem_read_result.is_mem_read_symbolic = memory_read_value.is_value_symbolic;
+		mem_read_result.read_size = read_size;
+		block_mem_reads_map.emplace(curr_instr_addr, mem_read_result);
+	}
+	else {
+		auto &mem_read_entry = block_mem_reads_map.at(curr_instr_addr);
+		mem_read_entry.memory_values.emplace_back(memory_read_value);
+		mem_read_entry.is_mem_read_symbolic |= memory_read_value.is_value_symbolic;
+		mem_read_entry.read_size += read_size;
+	}
+
+	// At this point the block's memory reads map has been rebuilt and we can propagate taint as before
+	auto &mem_read_result = block_mem_reads_map.at(curr_instr_addr);
 	if (curr_block_details.vex_lift_failed) {
-		if (mem_read_result.is_mem_read_symbolic || (symbolic_registers.size() > 0) || (block_symbolic_registers.size() > 0)) {
+		if (mem_read_result.is_mem_read_symbolic || (symbolic_registers.size() > 0)
+			|| (block_symbolic_registers.size() > 0) || (block_symbolic_temps.size() > 0)) {
 			// Either the memory value is symbolic or there are symbolic registers: thus, taint
 			// status of registers could change. But since VEX lift failed, the taint relations
 			// are not known and so we can't propagate taint. Stop concrete execution.
@@ -1611,23 +1750,31 @@ void State::propagate_taint_of_mem_read_instr_and_continue(const address_t instr
 		}
 		else {
 			// We cannot propagate taint since VEX lift failed and so we stop here. But, since
-			// there are no symbolic values, we do need need to propagate taint. Of course,
-			// the register slices cannot be updated but those won't be needed since this
-			// block will never be executed partially concretely and rest in VEX engine.
+			// there are no symbolic values, we do need need to propagate taint.
 			return;
 		}
 	}
-	if (mem_read_result.read_size != taint_engine_stop_mem_read_size) {
+	if (mem_read_result.read_size < taint_engine_stop_mem_read_size) {
 		// There are more bytes to be read by this instruction. We do not propagate taint until bytes are read
 		// Sometimes reads are split across multiple reads hooks in unicorn.
+		// Also, remember that a symbolic value has been partially read from memory so that even if the rest of the
+		// bytes to be read are concrete, taint will be propagated.
+		symbolic_read_in_progress = true;
 		return;
 	}
+	else if (mem_read_result.read_size > taint_engine_stop_mem_read_size) {
+		// Somehow the read result has read more bytes than read operation should according to the VEX statements.
+		// Likely a bug.
+		assert(false && "Memory read operation has read more bytes than expected. This should not happen!");
+	}
+
+	// Mark read as complete
+	symbolic_read_in_progress = false;
 
 	// There are no more pending reads at this instruction. Now we can propagate taint.
 	// This allows us to also handle cases when only some of the memory reads are symbolic: we treat all as symbolic
 	// and overtaint.
-	auto& block_taint_entry = block_taint_cache.at(curr_block_details.block_addr);
-	auto& instr_taint_data_entry = block_taint_entry.block_instrs_taint_data_map.at(instr_addr);
+	auto& instr_taint_data_entry = block_taint_entry.block_instrs_taint_data_map.at(curr_instr_addr);
 	if (mem_read_result.is_mem_read_symbolic || (symbolic_registers.size() > 0) || (block_symbolic_registers.size() > 0) ||
 	  block_symbolic_temps.size() > 0) {
 		if (block_taint_entry.has_unsupported_stmt_or_expr_type) {
@@ -1636,13 +1783,8 @@ void State::propagate_taint_of_mem_read_instr_and_continue(const address_t instr
 			stop(block_taint_entry.unsupported_stmt_stop_reason);
 			return;
 		}
-		compute_slice_of_instrs(instr_addr, instr_taint_data_entry);
-		propagate_taint_of_one_instr(instr_addr, instr_taint_data_entry);
+		propagate_taint_of_one_instr(curr_instr_addr, instr_taint_data_entry);
 	}
-	if (instr_slice_details_map.count(instr_addr) == 0) {
-		compute_slice_of_instrs(instr_addr, instr_taint_data_entry);
-	}
-	update_register_slice(instr_addr, instr_taint_data_entry);
 	if (!stopped) {
 		continue_propagating_taint();
 	}
@@ -1710,7 +1852,7 @@ void State::propagate_taint_of_one_instr(address_t instr_addr, const instruction
 				}
 			}
 			else if ((taint_sink.entity_type == TAINT_ENTITY_REG) && (taint_sink.reg_offset != arch_pc_reg_vex_offset())) {
-				// Mark register as concrete since none of it's dependencies are symbolic. Also update it's slice.
+				// Mark register as concrete since none of it's dependencies are symbolic.
 				mark_register_concrete(taint_sink.reg_offset, taint_sink.value_size);
 			}
 		}
@@ -1721,17 +1863,13 @@ void State::propagate_taint_of_one_instr(address_t instr_addr, const instruction
 		}
 	}
 	if (is_instr_symbolic) {
-		auto &instr_slice_details = instr_slice_details_map.at(instr_addr);
-		for (auto &reg: instr_slice_details.concrete_registers) {
-			auto reg_offset = reg.first;
-			auto reg_size = reg.second;
-			auto reg_val = block_start_reg_values.at(reg_offset);
-			reg_val.size = reg_size;
-			instr_details.reg_deps.insert(reg_val);
+		if (instr_details.has_symbolic_memory_dep) {
+			for (auto &mem_value: block_mem_reads_map.at(instr_addr).memory_values) {
+				if (mem_value.is_value_symbolic) {
+					instr_details.symbolic_mem_deps.emplace_back(std::make_pair(mem_value.address, mem_value.size));
+				}
+			}
 		}
-		instr_details.instr_deps.insert(instr_slice_details.dependent_instrs.begin(), instr_slice_details.dependent_instrs.end());
-		auto result = find_symbolic_mem_deps(instr_details);
-		instr_details.symbolic_mem_deps.insert(instr_details.symbolic_mem_deps.end(), result.begin(), result.end());
 		curr_block_details.symbolic_instrs.emplace_back(instr_details);
 	}
 	return;
@@ -1741,7 +1879,7 @@ instr_details_t State::compute_instr_details(address_t instr_addr, const instruc
 	instr_details_t instr_details;
 	instr_details.instr_addr = instr_addr;
 	if (instr_taint_entry.has_memory_read) {
-		auto mem_read_result = mem_reads_map.at(instr_addr);
+		auto mem_read_result = block_mem_reads_map.at(instr_addr);
 		if (!mem_read_result.is_mem_read_symbolic) {
 			instr_details.has_concrete_memory_dep = true;
 			instr_details.has_symbolic_memory_dep = false;
@@ -1771,61 +1909,29 @@ void State::start_propagating_taint(address_t block_address, int32_t block_size)
 		// We're not checking symbolic registers so no need to propagate taints
 		return;
 	}
-	if (this->block_taint_cache.find(block_address) == this->block_taint_cache.end()) {
-		// Compute and cache taint sink-source relations for this block
-		// Disable cross instruction optimization in IR so that dependencies of symbolic
-		// instructions can be computed correctly.
-		VexRegisterUpdates pxControl = VexRegUpdUnwindregsAtMemAccess;
-		std::unique_ptr<uint8_t[]> instructions(new uint8_t[block_size]);
-		address_t lift_address;
-
-		if ((arch == UC_ARCH_ARM) && is_thumb_mode()) {
-			lift_address = block_address | 1;
-		}
-		else {
-			lift_address = block_address;
-		}
-		uc_mem_read(this->uc, lift_address, instructions.get(), block_size);
-		VEXLiftResult *lift_ret = vex_lift(
-			this->vex_guest, this->vex_archinfo, instructions.get(), lift_address, 99, block_size, 1, 0, 1,
-			1, 0, pxControl
-		);
-
-		if ((lift_ret == NULL) || (lift_ret->size == 0)) {
-			// Failed to lift block to VEX.
-			if (symbolic_registers.size() > 0) {
-				// There are symbolic registers but VEX lift failed so we can't propagate taint
-				stop(STOP_VEX_LIFT_FAILED);
-			}
-			else {
-				// There are no symbolic registers so attempt to execute block. Mark block as VEX lift failed.
-				curr_block_details.vex_lift_failed = true;
-			}
+	if ((arch == UC_ARCH_ARM) && (block_taint_cache.find(block_address) == block_taint_cache.end())) {
+		// Block was not lifted and processed before. So it could end in syscall
+		curr_block_details.vex_lift_result = lift_block(block_address, block_size);
+		if ((curr_block_details.vex_lift_result == NULL) || (curr_block_details.vex_lift_result->size == 0)) {
+			// Failed to lift block to VEX. We don't execute the block because it could end in a syscall.
+			stop(STOP_VEX_LIFT_FAILED);
 			return;
 		}
-		if ((arch == UC_ARCH_ARM) && (lift_ret->irsb->jumpkind == Ijk_Sys_syscall)) {
+		if (curr_block_details.vex_lift_result->irsb->jumpkind == Ijk_Sys_syscall) {
 			// This block invokes a syscall. For now, such blocks are handled by VEX engine.
 			stop(STOP_SYSCALL_ARM);
 			return;
 		}
-		auto block_taint_entry = process_vex_block(lift_ret->irsb, block_address);
-		// Add entry to taint relations cache
-		block_taint_cache.emplace(block_address, block_taint_entry);
 	}
-	taint_engine_next_instr_address = block_address;
 	block_symbolic_temps.clear();
 	block_start_reg_values.clear();
-	for (auto &reg_instr_slice_entry: reg_instr_slice) {
-		// Clear slice for register
-		reg_instr_slice_entry.second.clear();
-		// Save value of all registers
+	// Save value of all registers in case some instruction touches symbolic data and needs to be re-executed
+	for (auto &reg_offset: vex_to_unicorn_map) {
 		register_value_t reg_value;
-		reg_value.offset = reg_instr_slice_entry.first;
+		reg_value.offset = reg_offset.first;
 		memset(reg_value.value, 0, MAX_REGISTER_BYTE_SIZE);
 		get_register_value(reg_value.offset, reg_value.value);
 		block_start_reg_values.emplace(reg_value.offset, reg_value);
-		// Reset the slice of register
-		reg_instr_slice_entry.second.clear();
 	}
 	for (auto &cpu_flag: cpu_flags) {
 		register_value_t flag_value;
@@ -1834,7 +1940,29 @@ void State::start_propagating_taint(address_t block_address, int32_t block_size)
 		get_register_value(cpu_flag.first, flag_value.value);
 		block_start_reg_values.emplace(flag_value.offset, flag_value);
 	}
-	propagate_taints();
+	if (symbolic_registers.size() != 0) {
+		if (block_taint_cache.find(block_address) == block_taint_cache.end()) {
+			// Compute and cache taint sink-source relations for this block since there are symbolic registers.
+			if (curr_block_details.vex_lift_result == NULL) {
+				curr_block_details.vex_lift_result = lift_block(block_address, block_size);
+				if ((curr_block_details.vex_lift_result == NULL) || (curr_block_details.vex_lift_result->size == 0)) {
+					// Failed to lift block to VEX.
+					if (symbolic_registers.size() > 0) {
+						// There are symbolic registers but VEX lift failed so we can't propagate taint
+						stop(STOP_VEX_LIFT_FAILED);
+					}
+					else {
+						// There are no symbolic registers so let's attempt to execute the block.
+						curr_block_details.vex_lift_failed = true;
+					}
+					return;
+				}
+			}
+			process_vex_block(curr_block_details.vex_lift_result->irsb, block_address);
+		}
+		taint_engine_next_instr_address = block_address;
+		propagate_taints();
+	}
 	return;
 }
 
@@ -1856,25 +1984,9 @@ void State::continue_propagating_taint() {
 	return;
 }
 
-std::vector<std::pair<address_t, uint64_t>> State::find_symbolic_mem_deps(const instr_details_t &instr) const {
-	std::vector<std::pair<address_t, uint64_t>> symbolic_mem_deps;
-	for (auto &dep_instr: instr.instr_deps) {
-		auto result = find_symbolic_mem_deps(dep_instr);
-		symbolic_mem_deps.insert(symbolic_mem_deps.end(), result.begin(), result.end());
-	}
-	if (instr.has_symbolic_memory_dep) {
-		for (auto &mem_value: mem_reads_map.at(instr.instr_addr).memory_values) {
-			if (mem_value.is_value_symbolic) {
-				symbolic_mem_deps.emplace_back(std::make_pair(mem_value.address, mem_value.size));
-			}
-		}
-	}
-	return symbolic_mem_deps;
-}
-
 void State::save_concrete_memory_deps(instr_details_t &instr) {
 	if (instr.has_concrete_memory_dep) {
-		archived_memory_values.emplace_back(mem_reads_map.at(instr.instr_addr).memory_values);
+		archived_memory_values.emplace_back(block_mem_reads_map.at(instr.instr_addr).memory_values);
 		instr.memory_values = &(archived_memory_values.back()[0]);
 		instr.memory_values_count = archived_memory_values.back().size();
 	}
@@ -1885,7 +1997,7 @@ void State::save_concrete_memory_deps(instr_details_t &instr) {
 	while (!instrs_to_process.empty()) {
 		auto &curr_instr = instrs_to_process.front();
 		if (curr_instr->has_concrete_memory_dep) {
-			archived_memory_values.emplace_back(mem_reads_map.at(curr_instr->instr_addr).memory_values);
+			archived_memory_values.emplace_back(block_mem_reads_map.at(curr_instr->instr_addr).memory_values);
 			curr_instr->memory_values = &(archived_memory_values.back()[0]);
 			curr_instr->memory_values_count = archived_memory_values.back().size();
 		}
@@ -1897,26 +2009,6 @@ void State::save_concrete_memory_deps(instr_details_t &instr) {
 	return;
 }
 
-void State::update_register_slice(address_t instr_addr, const instruction_taint_entry_t &curr_instr_taint_entry) {
-	instr_details_t instr_details = compute_instr_details(instr_addr, curr_instr_taint_entry);
-	for (auto &reg_entry: curr_instr_taint_entry.modified_regs) {
-		vex_reg_offset_t reg_offset = reg_entry.first;
-		auto reg_size = reg_entry.second.first;
-		if ((reg_offset == arch_pc_reg_vex_offset()) || is_symbolic_register(reg_offset, reg_size)) {
-			continue;
-		}
-		if (reg_instr_slice.find(reg_offset) == reg_instr_slice.end()) {
-			continue;
-		}
-		// Check if register's new value depends on it's previous value
-		if (!reg_entry.second.second) {
-			reg_instr_slice.at(reg_offset).clear();
-		}
-		reg_instr_slice.at(reg_offset).emplace_back(instr_details);
-	}
-	return;
-}
-
 bool State::is_block_exit_guard_symbolic() const {
 	auto& block_taint_entry = block_taint_cache.at(curr_block_details.block_addr);
 	auto block_exit_guard_taint_status = get_final_taint_status(block_taint_entry.exit_stmt_guard_expr_deps);
@@ -1981,56 +2073,7 @@ static void hook_mem_read(uc_engine *uc, uc_mem_type type, uint64_t address, int
 	// //LOG_D("mem_read [%#lx, %#lx] = %#lx", address, address + size);
 	//LOG_D("mem_read [%#lx, %#lx]", address, address + size);
 	State *state = (State *)user_data;
-	memory_value_t memory_read_value;
-	bool is_memory_value_symbolic;
-	address_t curr_instr_addr;
-
-	memory_read_value.reset();
-	memory_read_value.address = address;
-	memory_read_value.size = size;
-	if (!state->is_symbolic_taint_propagation_disabled()) {
-		// unicorn-engine doesn't update PC register in some cases and so it has incorrect value for current instruction
-		// address. XMM reads sometimes trigger the hook multiple times and there could be multiple reads in one instruction
-		// in some archs. So, we use the taint engine to determine instruction address correctly.
-		// See https://github.com/unicorn-engine/unicorn/issues/1312, https://github.com/unicorn-engine/unicorn/pull/1257
-		curr_instr_addr = state->get_taint_engine_stop_mem_read_instr_addr();
-	}
-	else {
-		// Symbolic taint propagation is disabled so we get current instruction address from unicorn
-		curr_instr_addr = state->get_instruction_pointer();
-	}
-	auto tainted = state->find_tainted(address, size);
-	if (tainted != -1)
-	{
-		if (state->is_symbolic_tracking_disabled()) {
-			// Symbolic register tracking is disabled but memory location has symbolic data.
-			// We switch to VEX engine then.
-			state->stop(STOP_SYMBOLIC_READ_SYMBOLIC_TRACKING_DISABLED);
-			return;
-		}
-		is_memory_value_symbolic = true;
-	}
-	else
-	{
-		is_memory_value_symbolic = false;
-		state->read_memory_value(address, size, memory_read_value.value, MAX_MEM_ACCESS_SIZE);
-	}
-	memory_read_value.is_value_symbolic = is_memory_value_symbolic;
-	auto mem_reads_map_entry = state->mem_reads_map.find(curr_instr_addr);
-	if (mem_reads_map_entry == state->mem_reads_map.end()) {
-		mem_read_result_t mem_read_result;
-		mem_read_result.memory_values.emplace_back(memory_read_value);
-		mem_read_result.is_mem_read_symbolic = is_memory_value_symbolic;
-		mem_read_result.read_size = size;
-		state->mem_reads_map.emplace(curr_instr_addr, mem_read_result);
-	}
-	else {
-		auto &mem_read_entry = state->mem_reads_map.at(curr_instr_addr);
-		mem_read_entry.memory_values.emplace_back(memory_read_value);
-		mem_read_entry.is_mem_read_symbolic |= is_memory_value_symbolic;
-		mem_read_entry.read_size += size;
-	}
-	state->propagate_taint_of_mem_read_instr_and_continue(curr_instr_addr);
+	state->propagate_taint_of_mem_read_instr_and_continue(address, size);
 	return;
 }
 
